#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Group Timetable caledar views.
"""
import urllib
import pytz
from datetime import datetime, timedelta
from time import strptime

from zope.publisher.browser import BrowserView
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.security.interfaces import Unauthorized
from zope.security.checker import canAccess
from zope.component import getUtility
from zope.i18n import translate

import schooltool.skin.flourish.page
from schooltool.app.browser.cal import CalendarViewBase
from schooltool.app.browser.cal import DailyCalendarView
from schooltool.app.browser.cal import WeeklyCalendarView
from schooltool.app.browser.cal import MonthlyCalendarView
from schooltool.app.browser.cal import YearlyCalendarView
from schooltool.app.browser.cal import EventForDisplay
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.calendar.utils import utcnow
from schooltool.resource.interfaces import IBookingCalendar
from schooltool.skin import flourish
from schooltool.person.interfaces import IPerson
from schooltool.app.cal import CalendarEvent
from schooltool.term.interfaces import IDateManager
from schooltool.app.browser import ViewPreferences

from schooltool.common import SchoolToolMessage as _


class BookingCalendarViewBase(CalendarViewBase):

    def getEvents(self, start_dt, end_dt):
        """Get a list of EventForDisplay objects for a selected time interval.

        `start_dt` and `end_dt` (datetime objects) are bounds (half-open) for
        the result.
        """
        for calendar, color1, color2 in self.getCalendars():
            for event in calendar.expand(start_dt, end_dt):
                yield EventForDisplay(event, self.request, color1, color2,
                                      calendar, self.timezone)

    def canAddEvents(self):
        """No one can add events to a booking calendar."""
        return False

    def canRemoveEvents(self):
        """No one can remove events from a booking calendar."""
        return False


class DailyBookingCalendarView(DailyCalendarView, BookingCalendarViewBase):
    pass


class WeeklyBookingCalendarView(WeeklyCalendarView, BookingCalendarViewBase):
    pass


class MonthlyBookingCalendarView(MonthlyCalendarView, BookingCalendarViewBase):
    pass


class YearlyBookingCalendarView(YearlyCalendarView, BookingCalendarViewBase):
    pass


class CalendarEventBookOneResourceView(BrowserView):
    """A view to book a resource to an event."""

    @property
    def timezone(self):
        prefs = ViewPreferences(self.request)
        return prefs.timezone

    def __call__(self):
        app = ISchoolToolApplication(None)
        person = IPerson(self.request.principal, None)
        if not person:
            raise Unauthorized("Only logged in users can book resources.")
        cal = ISchoolToolCalendar(person)
        if self.request.has_key('event_id'):
            event = cal.find(self.request['event_id'])
        else:
            start_date = self.request.get('start_date')
            start_time = self.request.get('start_time')
            title = self.request.get('title')
            start_datetime = "%s %s" % (start_date, start_time)
            start_datetime = datetime(*strptime(start_datetime,
                                                "%Y-%m-%d %H:%M")[0:6])
            start_datetime = self.timezone.localize(start_datetime)
            start_datetime = start_datetime.astimezone(pytz.UTC)
            duration = timedelta(seconds=int(self.request.get('duration')))
            event = CalendarEvent(dtstart = start_datetime,
                                  duration = duration,
                                  title = title)
            cal.addEvent(event)

        if event:
            resource = app["resources"].get(self.request['resource_id'])
            if resource is not None:
                resource_calendar = ISchoolToolCalendar(resource)
                if not canAccess(resource_calendar, "addEvent"):
                    raise Unauthorized("You don't have the right to"
                                       " book this resource!")
                event.bookResource(resource)
        self.request.response.redirect(self.nextURL(event))

    def cancelURL(self, event, back_url):
        calURL = absoluteURL(event.__parent__, self.request)
        cancel_url = (
            calURL + '/delete.html' +
            '?event_id=%s' % event.unique_id +
            '&date=%s' % event.dtstart.strftime("%Y-%m-%d") +
            '&back_url=%s' % back_url +
            '&DELETE=Delete')
        return cancel_url

    def nextURL(self, event):
        """Return the URL to be displayed after the add operation."""

        app = ISchoolToolApplication(None)
        resource = app["resources"].get(self.request['resource_id'])

        back_url = self.request.get('next_url', '')
        if resource is not None and not back_url:
            back_url = urllib.quote(absoluteURL(
                ISchoolToolCalendar(resource), self.request))

        cancel_url = self.cancelURL(event, back_url)

        url = "%s/edit.html?back_url=%s&cancel_url=%s" % (
            absoluteURL(event, self.request),
            back_url,
            urllib.quote(cancel_url))
        return url



class FlourishCalendarEventBookOneResourceView(
    CalendarEventBookOneResourceView):

    def render(self, *args, **kw):
        return ''

    def update(self):
        self.__call__()

    def cancelURL(self, event, back_url):
        calURL = absoluteURL(event.__parent__, self.request)
        cancel_url = (
            calURL + '/delete-temp.html' +
            '?event_id=%s' % event.unique_id +
            '&date=%s' % event.dtstart.strftime("%Y-%m-%d") +
            '&back_url=%s' % back_url +
            '&DELETE=Delete')
        return cancel_url


class BookResourceLink(flourish.page.LinkViewlet):


    def nextURL(self):
        return urllib.quote(absoluteURL(self.context, self.request))

    @property
    def url(self):
        rc = ISchoolToolApplication(None)['resources']
        booking_calendar = IBookingCalendar(rc)
        url = absoluteURL(booking_calendar, self.request)
        url = "%s/book_one_resource.html?resource_id=%s" % (
            url, self.context.__name__)
        today = getUtility(IDateManager).today
        hour = utcnow().hour
        duration = 3600
        url = "%s&start_date=%s&start_time=%s:00&duration=%s&title=%s" % (
            url, today.isoformat(), hour, duration,
            translate(_("Unnamed Event"), context=self.request))
        url = "%s&next_url=%s" % (url, self.nextURL())
        return url



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Configuration views for SchoolTool security policy.
"""

from zope.publisher.browser import BrowserView
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.securitypolicy.interfaces import IAccessControlCustomisations

from schooltool.common import SchoolToolMessage as _


class AccessControlView(BrowserView):

    def settings(self):
        """Returns a list of access control customisation settings."""
        app = ISchoolToolApplication(None)
        customisations = IAccessControlCustomisations(app)
        return list(customisations)

    def update(self):
        prefix = 'setting.'
        if 'UPDATE_SUBMIT' in self.request:
            for setting in self.settings():
                val = self.request.get(prefix + setting.key, 'False')
                setting.setValue(bool(val != 'False'))
        elif 'CANCEL' in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)


class FlourishAccessControlView(AccessControlView):

    fieldset_titles = {
        'everyone_can_view_section_info': _('Section information'),
        'instructors_can_edit_section': _('Section information'),
        'everyone_can_view_person_list': _('List of accounts'),
        'persons_can_set_their_preferences': _('User calendars'),
        'persons_can_change_their_passwords': _('User passwords'),
        'everyone_can_view_group_list': _('List of groups'),
        'everyone_can_view_group_info': _('Group information'),
        'everyone_can_view_group_calendar': _('View group calendars'),
        'members_can_edit_group_calendar': _('Edit group calendars'),
        'everyone_can_view_resource_list': _('List of resources'),
        'everyone_can_view_resource_info': _('Resource information'),
        'everyone_can_view_resource_calendar': _('Resource calendars'),
        'instructors_can_schedule_sections': _('Schedules'),
        'administration_can_grade_students': _('Gradebook'),
        'administration_can_grade_journal': _('Journal'),
        }

    def fieldsets(self):
        for setting in self.settings():
            info = {'legend': self.fieldset_titles[setting.key],
                    'setting': setting}
            yield info

    def update(self):
        prefix = 'setting.'
        self.status = None
        if 'UPDATE_SUBMIT' in self.request:
            for setting in self.settings():
                val = self.request.get(prefix + setting.key, 'False')
                setting.setValue(bool(val != 'False'))
            url = absoluteURL(self.context, self.request) + '/security.html'
            self.request.response.redirect(url)
        elif 'CANCEL' in self.request:
            url = absoluteURL(self.context, self.request) + '/security.html'
            self.request.response.redirect(url)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool application views.
"""
import csv
import transaction

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.publisher.browser import BrowserView

import schooltool.skin.flourish.page
from schooltool.common import SchoolToolMessage as _
from schooltool.app.app import SimpleNameChooser
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.skin import flourish


class ImportErrorCollection(object):
    """A simple container for import errors.

    This class just holds errors that occur in the CSVImporter class so they
    can be dealt with by the CSVImportView class.
    """

    def __init__(self):
        self.generic = []
        self.fields = []

    def anyErrors(self):
        return bool(self.generic or self.fields)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.__dict__)


class BaseCSVImportView(BrowserView):

    __used_for__ = ISchoolToolApplication

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        self.errors = []
        self.success = []

    def nextURL(self):
        return None

    @property
    def mode(self):
        mode = self.request.get('csv-input-mode')
        if (not mode and
            not self.request.get('csvfile') and
            self.request.get('csvtext')):
            return 'text'
        return mode or 'upload'

    def update(self):
        if "UPDATE_SUBMIT" not in self.request:
            return

        charset = self.getCharset()
        if charset is None:
            return

        csvfile = self.request.get('csvfile', '')
        if csvfile:
            csvfile = csvfile.read()

        csvtext = self.request.get('csvtext', '').strip()

        mode = self.mode

        if (mode == 'upload' and not csvfile or
            mode == 'text' and not csvtext):
            self.errors.append(_('No data provided'))
            return

        if mode == 'upload' and csvfile:
            try:
                unicode(csvfile, charset)
            except UnicodeError:
                self.errors.append(_('Could not convert data to Unicode'
                                     ' (incorrect charset?).'))
                return

        self.importer = self.importer_class(self.context, charset)
        ok = True

        if mode == 'upload':
            ok = self.importer.importFromCSV(csvfile)
            if ok:
                self.success.append(_("CSV file imported successfully."))
            else:
                self.errors.append(_("Failed to import CSV file"))
                self._presentErrors(self.importer.errors)
        elif mode == 'text':
            self.csvtext = csvtext
            default_charset = "utf-8"
            self.importer.charset = default_charset
            ok = self.importer.importFromCSV(csvtext.encode(default_charset))
            if ok:
                self.success.append(_("CSV text imported successfully."))
            else:
                self.errors.append(_("Failed to import CSV text"))
                self._presentErrors(self.importer.errors)

        if self.success:
            next_url = self.nextURL()
            if next_url:
                self.request.response.redirect(next_url)

    def _presentErrors(self, err):
        """Add any errors in our ErrorCollection to the view errors.
        """
        if err.generic:
            self.errors.extend(err.generic)

        if err.fields:
            self.errors.extend(err.fields)

    def getCharset(self):
        """Return the charset (as a string) that was specified in the request.

        Updates self.errors and returns None if the charset was not specified
        or if it is invalid.
        """
        charset = self.request['charset']
        if charset == 'other':
            charset = self.request['other_charset']
        if not charset:
            self.errors.append(_("No charset specified"))
            return
        try:
            unicode(' ', charset)
        except LookupError:
            self.errors.append(_("Unknown charset"))
            return
        return charset


class FlourishBaseCSVImportView(flourish.page.Page, BaseCSVImportView):
    content_template = ViewPageTemplateFile('templates/f_csvimport.pt')
    __init__ = BaseCSVImportView.__init__
    update = BaseCSVImportView.update

    def update(self):
        if "UPDATE_CANCEL" in self.request:
            self.request.response.redirect(self.nextURL())
            return
        return BaseCSVImportView.update(self)


class BaseCSVImporter(object):
    """A base class for CSV parsers and importers.

    Subclasses should define the createAndAdd method.
    """

    def __init__(self, container, charset=None):
        self.container = container
        self.errors = ImportErrorCollection()
        self.charset = charset
        self.chooser = SimpleNameChooser(container)

    def parseCSVRows(self, rows):
        """Parse rows (a list of strings) in CSV format.

        Returns a list of rows as lists.

        rows must be in the encoding specified during construction of
        BaseCSVImportView; the returned values are in unicode.

        If the provided data is invalid, self.errors.generic will be updated
        and None will be returned.
        """
        result = []
        reader = csv.reader(rows)
        line = 0
        try:
            while True:
                line += 1
                values = [v.strip() for v in reader.next()]
                if self.charset:
                    values = [unicode(v, self.charset) for v in values]
                # Remove trailing empty cells.
                while values and not values[-1].strip():
                    del values[-1]
                result.append(values)
        except StopIteration:
            return result
        except csv.Error:
            error_msg = _("Error in CSV data, line ${line_no}",
                          mapping={'line_no': line})
            self.errors.generic.append(error_msg)
        except UnicodeError:
            error_msg = _("Conversion to unicode failed in line ${line_no}",
                          mapping={'line_no': line})
            self.errors.generic.append(error_msg)

    def importFromCSV(self, csvdata):
        """Import objects from CSV data.

        Should not throw exceptions, but will set self.*error attributes.
        Returns True on success.  If False is returned, it means that at least
        one of attributes of self.errors have been set, and that no changes to
        the database have been applied.
        """
        rows = self.parseCSVRows(csvdata.strip().splitlines())
        if rows is None:
            return False

        for dry_run in [True, False]:
            savepoint = transaction.savepoint()
            for rowdata in rows:
                self.createAndAdd(rowdata, dry_run)

            if self.errors.anyErrors():
                savepoint.rollback()
                return False

        return True

    def createAndAdd(self, obj, dry_run=True):
        """Create object and add to container.

        If dry_run is True, don't actually do anything, just validate the data.

        This should be defined in the subclass.
        """
        raise NotImplementedError("Please override this method in subclasses")





#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 36.
"""
import transaction

from schooltool.testing.mock import ModulesSnapshot
from schooltool.generations.evolve36.helper import BuildContext
from schooltool.generations.evolve36.timetable_builders import \
    SchoolTimetablesBuilder
from schooltool.generations.evolve36.schedule_builders import \
    AppSchedulesBuilder
from schooltool.generations.evolve36.calendar_builders import \
    AppTimetableCalendarBuilder


def evolveTimetables(app):
    # This import holds references to substitute classes
    # so that they can be pickled after modules are restored.
    from schooltool.generations.evolve36 import model

    modules = ModulesSnapshot()

    modules.mock_module('schooltool.timetable')
    modules.mock(model.substitutes)

    builders = [
        SchoolTimetablesBuilder(),
        AppSchedulesBuilder(),
        AppTimetableCalendarBuilder(),
        ]

    for builder in builders:
        builder.read(app, BuildContext())

    modules.restore()

    for builder in builders:
        builder.clean(app, BuildContext())

    result = BuildContext()
    for builder in builders:
        built = builder.build(app, BuildContext(shared=result))
        result.update(built)

    # Pickle stuff
    transaction.savepoint()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Selenium Functional Testing Utilities for export.
"""

import os

from schooltool.testing.selenium import SeleniumLayer

dir = os.path.abspath(os.path.dirname(__file__))
filename = os.path.join(dir, 'stesting.zcml')

export_selenium_layer = SeleniumLayer(filename,
                                     __name__,
                                     'export_selenium_layer')

def registerSeleniumSetup():
    try:
        import selenium
    except ImportError:
        return
    from schooltool.testing import registry
    import schooltool.testing.selenium

    def importXLS(browser, filename):
        browser.query.link('School').click()
        browser.query.link('Import XLS').click()
        if filename:
            import pkg_resources
            elem = browser.query.name('xlsfile')
            browser.query.name('xlsfile').type(filename)
        page = browser.query.tag('html')
        browser.query.button('Submit').click()
        browser.wait(lambda: page.expired)

    registry.register('SeleniumHelpers',
        lambda: schooltool.testing.selenium.registerBrowserUI(
            'import_xls', importXLS))

registerSeleniumSetup()
del registerSeleniumSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2004 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.common
"""

import unittest
import doctest
import datetime

from zope.traversing.testing import setUp


class TestHelpers(unittest.TestCase):

    # parse_date has a doctest

    def test_parse_datetime(self):
        from schooltool.common import parse_datetime
        dt = datetime.datetime
        valid_dates = (
            ("2000-01-01 00:00:00", dt(2000, 1, 1, 0, 0, 0, 0)),
            ("2000-01-01 00:00:00.000000", dt(2000, 1, 1, 0, 0, 0, 0)),
            ("2000-01-01T00:00:00", dt(2000, 1, 1, 0, 0, 0, 0)),
            ("2005-12-23 11:22:33", dt(2005, 12, 23, 11, 22, 33)),
            ("2005-12-23T11:22:33", dt(2005, 12, 23, 11, 22, 33)),
            ("2005-12-23T11:22:33.4", dt(2005, 12, 23, 11, 22, 33, 400000)),
            ("2005-12-23T11:22:33.456789", dt(2005, 12, 23, 11, 22, 33,
                                              456789)),
        )
        for s, d in valid_dates:
            result = parse_datetime(s)
            self.assertEquals(result, d,
                              "parse_datetime(%r) returned %r" % (s, result))
        invalid_dates = (
            "2000/01/01",
            "2100-02-29 00:00:00",
            "2005-12-23 11:22:33 "
        )
        for s in invalid_dates:
            try:
                result = parse_datetime(s)
            except ValueError:
                pass
            else:
                self.fail("parse_datetime(%r) did not raise" % s)


def test_suite():
    optionflags = (doctest.ELLIPSIS
                   | doctest.REPORT_NDIFF
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite('schooltool.common',
                                       optionflags=optionflags))
    suite.addTest(doctest.DocTestSuite('schooltool.common.inlinept',
                                       optionflags=optionflags,
                                       setUp=lambda test: setUp()))
    suite.addTest(unittest.makeSuite(TestHelpers))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for group views.
"""

import unittest
import doctest

from zope.interface import directlyProvides
from zope.publisher.browser import TestRequest
from zope.traversing.interfaces import IContainmentRoot
from zope.component import provideAdapter

from schooltool.app.browser.testing import setUp, tearDown
from schooltool.testing import setup


def doctest_GroupListView():
    r"""Test for GroupListView

    We will need a volunteer for this test:

        # XXX: Should use stub
        >>> from schooltool.person.person import Person
        >>> person = Person(u'ignas')

    One requirement: the person has to know where he is.

        >>> app = setup.setUpSchoolToolSite()
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> provideAdapter(lambda context: app, (None,), ISchoolToolApplication)
        >>> app['persons']['ignas'] = person

    We will be testing the person's awareness of the world, so we will
    create some (empty) groups.

        >>> from schooltool.group.group import Group
        >>> from schooltool.group.group import GroupContainer
        >>> from zope.component import provideAdapter
        >>> from schooltool.group.interfaces import IGroupContainer
        >>> gc = GroupContainer()
        >>> provideAdapter(lambda x: gc, adapts=[None], provides=IGroupContainer)
        >>> world = gc['the_world'] = Group("Others")
        >>> etria = gc['etria'] = Group("Etria")
        >>> pov = gc['pov'] = Group("PoV")
        >>> canonical = gc['canonical'] = Group("Canonical")
        >>> ms = gc['ms'] = Group("The Enemy")

    And we need a table formatter to display our groups:

        >>> from schooltool.table.table import SchoolToolTableFormatter
        >>> from schooltool.table.interfaces import ITableFormatter
        >>> from zope.publisher.interfaces.browser import IBrowserRequest
        >>> provideAdapter(SchoolToolTableFormatter, (None, IBrowserRequest), ITableFormatter)

    Let's create a view for a person:

        >>> from schooltool.group.browser.group import GroupListView
        >>> request = TestRequest()
        >>> view = GroupListView(person, request)
        >>> view.filter = lambda l: l

    Rendering the view does no harm:

        >>> view.update()

    First, all groups the person is not a member of should be listed:

        >>> sorted([g.title for g in view.getAvailableItems()])
        ['Canonical', 'Etria', 'Others', 'PoV', 'The Enemy']

    As well as all groups the person is currently a member of:

        >>> sorted([g.title for g in view.getSelectedItems()])
        []

    Let's tell the person to join PoV:

        >>> request = TestRequest()
        >>> request.form = {'add_item.pov': 'on', 'ADD_ITEMS': 'Apply'}
        >>> view = GroupListView(person, request)
        >>> view.filter = lambda l: l
        >>> view.update()

    He should have joined:

        >>> [group.title for group in person.groups]
        ['PoV']

    Had we decided to make the guy join Etria but then changed our mind:

        >>> request = TestRequest()
        >>> request.form = {'remove_item.pov': 'on', 'add_group.etria': 'on',
        ...                 'CANCEL': 'Cancel'}
        >>> view = GroupListView(person, request)
        >>> view.filter = lambda l: l
        >>> view.update()

    Nothing would have happened!

        >>> [group.title for group in person.groups]
        ['PoV']

    Yet we would find ourselves in the person info page:

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/persons/ignas'

    Finally, let's remove him out of PoV for a weekend and add him
    to The World.

        >>> request = TestRequest()
        >>> request.form = {'remove_item.pov': 'on', 'REMOVE_ITEMS': 'Apply'}
        >>> view = GroupListView(person, request)
        >>> view.filter = lambda l: l
        >>> view.update()

    Mission successful:

        >>> [group.title for group in person.groups]
        []

    """


def doctest_MemberListView():
    r"""Test for MemberListView

    We will be (ab)using a group and three test subjects:

        >>> from schooltool.group.group import Group
        >>> pov = Group('PoV')

        # XXX: Use stub implementation
        >>> from schooltool.person.person import Person
        >>> gintas = Person('gintas', 'Gintas')
        >>> ignas = Person('ignas', 'Ignas')
        >>> alga = Person('alga', 'Albertas')

    We need these objects to live in an application:

        >>> app = setup.setUpSchoolToolSite()
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> provideAdapter(lambda context: app, (None,), ISchoolToolApplication)

        >>> from schooltool.group.group import GroupContainer
        >>> from zope.component import provideAdapter
        >>> from schooltool.group.interfaces import IGroupContainer
        >>> gc = GroupContainer()
        >>> gc.__parent__, gc.__name__ = app, 'groups'
        >>> provideAdapter(lambda x: gc, adapts=[None], provides=IGroupContainer)

        >>> gc['pov'] = pov
        >>> app['persons']['gintas'] = gintas
        >>> app['persons']['ignas'] = ignas
        >>> app['persons']['alga'] = alga

    And we need a table formatter to display our persons:

        >>> from schooltool.table.table import SchoolToolTableFormatter
        >>> from schooltool.table.interfaces import ITableFormatter
        >>> from zope.publisher.interfaces.browser import IBrowserRequest
        >>> provideAdapter(SchoolToolTableFormatter, (None, IBrowserRequest), ITableFormatter)

    Let's create a view for our group:

        >>> from schooltool.group.browser.group import MemberViewPersons
        >>> request = TestRequest()
        >>> view = MemberViewPersons(pov, request)
        >>> view.filter = lambda l: l

    Rendering the view does no harm:

        >>> view.update()

    First, all persons should be listed (the page template puts them in
    alphabetical order later):

        >>> sorted([g.title for g in view.getAvailableItems()])
        ['Albertas', 'Gintas', 'Ignas']

    Let's make Ignas a member of PoV:

        >>> request = TestRequest()
        >>> request.form = {'add_item.ignas': 'on', 'ADD_ITEMS': 'Apply'}
        >>> view = MemberViewPersons(pov, request)
        >>> view.filter = lambda l: l
        >>> view.update()

    He should have joined:

        >>> sorted([person.title for person in pov.members])
        ['Ignas']

    We can cancel an action if we want to:

        >>> request = TestRequest()
        >>> request.form = {'add_item.gintas': 'on', 'CANCEL': 'Cancel'}
        >>> view = MemberViewPersons(pov, request)
        >>> view.filter = lambda l: l
        >>> view.update()
        >>> sorted([person.title for person in pov.members])
        ['Ignas']
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/groups/pov'

    Let's remove Ignas from PoV (he went home early today);

        >>> request = TestRequest()
        >>> request.form = {'remove_item.ignas': 'on', 'REMOVE_ITEMS': 'Apply'}
        >>> view = MemberViewPersons(pov, request)
        >>> view.filter = lambda l: l
        >>> view.update()

    and add Albert, who came in late and has to work after-hours:

        >>> request = TestRequest()
        >>> request.form = {'add_item.alga': 'on', 'ADD_ITEMS': 'Apply'}
        >>> view = MemberViewPersons(pov, request)
        >>> view.filter = lambda l: l
        >>> view.update()

    Mission accomplished:

        >>> sorted([person.title for person in pov.members])
        ['Albertas']

    Click 'Done' when we are finished and we go back to the group view

        >>> request = TestRequest()
        >>> request.form = {'CANCEL': 'Cancel'}
        >>> view = MemberViewPersons(pov, request)
        >>> view.filter = lambda l: l
        >>> view.update()
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/groups/pov'

    TODO: check resource view

    """


def doctest_GroupView():
    r"""Test for GroupView

    Let's create a view for a group:

        >>> from schooltool.group.browser.group import GroupView
        >>> from schooltool.group.group import Group
        >>> group = Group()
        >>> request = TestRequest()
        >>> view = GroupView(group, request)

    Let's relate some objects to our group:

        >>> from schooltool.person.person import Person
        >>> group.members.add(Person(title='First'))
        >>> group.members.add(Person(title='Last'))
        >>> group.members.add(Person(title='Intermediate'))

    Only persons whose title we can see are in the list, so we must
    define an all alowing security checker:

        >>> from zope.security.checker import defineChecker, Checker
        >>> defineChecker(Person,
        ...               Checker({'title': 'zope.Public'},
        ...                       {'title': 'zope.Public'}))

        >>> sorted([person.title for person in view.getPersons()])
        ['First', 'Intermediate', 'Last']

    """


def doctest_GroupAddView():
    r"""Test for GroupAddView

    Adding views in Zope 3 are somewhat unobvious.  The context of an adding
    view is a view named '+' and providing IAdding.

        >>> class AddingStub:
        ...     pass
        >>> context = AddingStub()

    The container to which items will actually be added is accessible as the
    `context` attribute

        >>> from schooltool.group.group import GroupContainer
        >>> container = GroupContainer()
        >>> context.context = container

    ZCML configuration adds some attributes to GroupAddView, namely `schema`,
    'fieldNames', and `_factory`.

        >>> from schooltool.group.browser.group import GroupAddView
        >>> from schooltool.group.interfaces import IGroup
        >>> from schooltool.group.group import Group
        >>> class GroupAddViewForTesting(GroupAddView):
        ...     schema = IGroup
        ...     fieldNames = ('title', 'description')
        ...     _factory = Group

    We can now finally create the view:

        >>> request = TestRequest()
        >>> view = GroupAddViewForTesting(context, request)

    The `nextURL` method tells Zope 3 where you should be redirected after
    successfully adding a group.  We will pretend that `container` is located
    at the root so that absoluteURL(container) returns 'http://127.0.0.1'.

        >>> directlyProvides(container, IContainmentRoot)
        >>> view.nextURL()
        'http://127.0.0.1'

    We can cancel an action if we want to:

        >>> request = TestRequest()
        >>> request.form = {'CANCEL': 'Cancel'}
        >>> view = GroupAddViewForTesting(context, request)
        >>> view.update()
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

    If 'CANCEL' is not present in the request, the view calls inherited
    'update'.  We will use a trick and set update_status to some value to
    short-circuit AddView.update().

        >>> request = TestRequest()
        >>> request.form = {'field.title': 'a_group',
        ...                 'UPDATE_SUBMIT': 'Add'}
        >>> view = GroupAddViewForTesting(context, request)
        >>> view.update_status = 'Just checking'
        >>> view.update()
        'Just checking'

    """


def doctest_GroupEditView():
    r"""Test for GroupEditView

    Let's create a view for editing a group:

        >>> from schooltool.group.browser.group import GroupEditView
        >>> from schooltool.group.group import Group
        >>> from schooltool.group.interfaces import IGroup
        >>> group = Group()
        >>> directlyProvides(group, IContainmentRoot)
        >>> request = TestRequest()

        >>> class TestGroupEditView(GroupEditView):
        ...     schema = IGroup
        ...     fieldNames = ('title', 'description')
        ...     _factory = Group

        >>> view = TestGroupEditView(group, request)

    We should not get redirected if we did not click on apply button:

        >>> request = TestRequest()
        >>> view = TestGroupEditView(group, request)
        >>> view.update()
        ''
        >>> request.response.getStatus()
        599

    After changing name of the group you should get redirected to the group
    list:

        >>> request = TestRequest()
        >>> request.form = {'UPDATE_SUBMIT': 'Apply',
        ...                 'field.title': u'new_title'}
        >>> view = TestGroupEditView(group, request)
        >>> view.update()
        u'Updated on ${date_time}'
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

        >>> group.title
        u'new_title'

    Even if the title has not changed you should get redirected to the group
    list:

        >>> request = TestRequest()
        >>> request.form = {'UPDATE_SUBMIT': 'Apply',
        ...                 'field.title': u'new_title'}
        >>> view = TestGroupEditView(group, request)
        >>> view.update()
        ''
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

        >>> group.title
        u'new_title'

    We should not get redirected if there were errors:

        >>> request = TestRequest()
        >>> request.form = {'UPDATE_SUBMIT': 'Apply',
        ...                 'field.title': u''}
        >>> view = TestGroupEditView(group, request)
        >>> view.update()
        u'An error occurred.'
        >>> request.response.getStatus()
        599

        >>> group.title
        u'new_title'

    We can cancel an action if we want to:

        >>> request = TestRequest()
        >>> request.form = {'CANCEL': 'Cancel'}
        >>> view = TestGroupEditView(group, request)
        >>> view.update()
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1'

    """


def doctest_GroupCSVImporter():
    r"""Tests for GroupCSVImporter.

    Create a group container and an importer

        >>> from schooltool.group.browser.csvimport import GroupCSVImporter
        >>> from schooltool.group.group import GroupContainer
        >>> container = GroupContainer()
        >>> importer = GroupCSVImporter(container, None)

    Import some sample data

        >>> csvdata='''Group 1, Group 1 Description
        ... Group2
        ... Group3, Group 3 Description, Some extra data\n\n\n'''
        >>> importer.importFromCSV(csvdata)
        True

    Check that the groups exist

        >>> [group for group in container]
        [u'group-1', u'group2', u'group3']

    Check that descriptions were imported properly

        >>> [group.description for group in container.values()]
        ['Group 1 Description', '', 'Group 3 Description']

    """


def doctest_GroupCSVImportView():
    r"""Tests for GroupCSVImportView

    We'll create a group csv import view

        >>> from schooltool.group.browser.csvimport import \
        ...     GroupCSVImportView
        >>> from schooltool.group.group import GroupContainer
        >>> from zope.publisher.browser import TestRequest
        >>> container = GroupContainer()
        >>> request = TestRequest()

    Now we'll try a text import.  Note that the description is not required

        >>> request.form = {
        ...     'csvtext' : u'A Group, The best Group\nAnother Group\nEspañol, Spanish Group\n\n\n',
        ...     'charset' : 'UTF-8',
        ...     'UPDATE_SUBMIT': 1}
        >>> view = GroupCSVImportView(container, request)
        >>> view.update()
        >>> [group for group in container]
        [u'a-group', u'another-group', u'espa\xe3ol']

    If no data is provided, we naturally get an error

        >>> request.form = {'charset' : 'UTF-8', 'UPDATE_SUBMIT': 1}
        >>> view = GroupCSVImportView(container, request)
        >>> view.update()
        >>> view.errors
        [u'No data provided']

    We also get an error if a line starts with a comma (no title)

        >>> request.form = {'csvtext' : ", No title provided here",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = GroupCSVImportView(container, request)
        >>> view.update()
        >>> view.errors
        [u'Failed to import CSV text', u'Titles may not be empty']

    """


def doctest_GroupMemberCSVImporter():
    r"""Tests for GroupMemberCSVImporter.

    First we need to set up some persons:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.person.person import Person
        >>> school = setup.setUpSchoolToolSite()
        >>> provideAdapter(lambda context: school, (None,), ISchoolToolApplication)
        >>> persons = school['persons']
        >>> directlyProvides(school, IContainmentRoot)
        >>> smith = persons['smith'] = Person('smith', 'John Smith')
        >>> [group.title for group in smith.groups]
        []
        >>> jones = persons['jones'] = Person('jones', 'Sally Jones')
        >>> [group.title for group in jones.groups]
        []
        >>> stevens = persons['stevens'] = Person('stevens', 'Bob Stevens')
        >>> [group.title for group in stevens.groups]
        []

    Create a group and an importer

        >>> from schooltool.group.browser.csvimport import GroupMemberCSVImporter
        >>> from schooltool.group.group import Group
        >>> group = Group('Group title', 'Group description')
        >>> [person.username for person in group.members]
        []
        >>> importer = GroupMemberCSVImporter(group, None)

    Import some sample data

        >>> csvdata='''smith
        ... stevens\n\n\n'''
        >>> importer.importFromCSV(csvdata)
        True

    Check that the persons were added to the group members:

        >>> [person.username for person in group.members]
        ['smith', 'stevens']
        >>> [group.title for group in smith.groups]
        ['Group title']
        >>> [group.title for group in jones.groups]
        []
        >>> [group.title for group in stevens.groups]
        ['Group title']

    Create another group and another importer

        >>> another_group = Group('Another group', 'Another description')
        >>> [person.username for person in another_group.members]
        []
        >>> another_importer = GroupMemberCSVImporter(another_group, None)

    Import some more data

        >>> csvdata='''stevens
        ... jones\n\n\n'''
        >>> another_importer.importFromCSV(csvdata)
        True

    Check that the persons were added to the another group members:

        >>> [person.username for person in another_group.members]
        ['stevens', 'jones']
        >>> [group.title for group in smith.groups]
        ['Group title']
        >>> [group.title for group in jones.groups]
        ['Another group']
        >>> [group.title for group in stevens.groups]
        ['Group title', 'Another group']

    """


def doctest_GroupMemberCSVImportView():
    r"""Tests for GroupMemberCSVImportView

    First we need to set up some persons:

        >>> from zope.i18n import translate
        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.person.person import Person
        >>> school = setup.setUpSchoolToolSite()
        >>> provideAdapter(lambda context: school, (None,), ISchoolToolApplication)
        >>> persons = school['persons']
        >>> directlyProvides(school, IContainmentRoot)
        >>> smith = persons['smith'] = Person('smith', 'John Smith')
        >>> jones = persons['jones'] = Person('jones', 'Sally Jones')
        >>> stevens = persons['stevens'] = Person('stevens', 'Bob Stevens')

    We'll create a group member csv import view

        >>> from schooltool.group.browser.csvimport import \
        ...      GroupMemberCSVImportView
        >>> from schooltool.group.group import Group
        >>> from zope.publisher.browser import TestRequest
        >>> group = Group('Group title', 'Group description')
        >>> request = TestRequest()

    Now we'll try a text import.

        >>> request.form = {
        ...     'csvtext' : 'stevens\n\n\n',
        ...     'charset' : 'UTF-8',
        ...     'UPDATE_SUBMIT': 1}
        >>> view = GroupMemberCSVImportView(group, request)
        >>> view.update()
        >>> [person.username for person in group.members]
        ['stevens']

    If no data is provided, we naturally get an error

        >>> request.form = {'charset' : 'UTF-8', 'UPDATE_SUBMIT': 1}
        >>> view.update()
        >>> view.errors
        [u'No data provided']

    We also get an error if a line doesn't have a username

        >>> request.form = {'csvtext' : " ,stevens\njones,Sally\n\n",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = GroupMemberCSVImportView(group, request)
        >>> view.update()
        >>> view.errors
        [u'Failed to import CSV text', u'User names must not be empty.']

    Or if the username is not in the persons container

        >>> request.form = {'csvtext' : "foobar\nstevens\njones",
        ...                 'charset' : 'UTF-8',
        ...                 'UPDATE_SUBMIT': 1}
        >>> view = GroupMemberCSVImportView(group, request)
        >>> view.update()
        >>> [translate(error) for error in view.errors]
        [u'Failed to import CSV text', u'"foobar" is not a valid username.']

    """


def doctest_GroupsViewlet():
    r"""Test for GroupsViewlet

    Adapters:

        >>> from zope.ucol.localeadapter import LocaleCollator
        >>> from zope.component import provideAdapter, provideUtility
        >>> from zope.i18n.interfaces.locales import ICollator
        >>> from zope.i18n.interfaces.locales import ILocale
        >>> from schooltool.course.section import PersonInstructorAdapter
        >>> from schooltool.course.section import PersonLearnerAdapter
        >>> from schooltool.group.interfaces import IGroupContainer
        >>> from schooltool.schoolyear.interfaces import ISchoolYear
        >>> provideAdapter(LocaleCollator, [ILocale], ICollator)
        >>> provideAdapter(PersonInstructorAdapter)
        >>> provideAdapter(PersonLearnerAdapter)
        >>> class DummySchoolYear(object):
        ...     title = '2010'
        ...     first = 'date'
        >>> year2010 = DummySchoolYear()
        >>> def DummySchoolYearSectionAdapter(section):
        ...     return year2010
        >>> provideAdapter(DummySchoolYearSectionAdapter, [IGroupContainer],
        ...                ISchoolYear)

    Let's create a viewlet for a person's groups:

        >>> from schooltool.group.browser.group import GroupsViewlet
        >>> from schooltool.person.person import Person

        >>> school = setup.setUpSchoolToolSite()
        >>> persons = school['persons']
        >>> persons['student'] = student = Person("Student")

    We need to allow access to group's titles.

        >>> from zope.security.checker import defineChecker, Checker
        >>> from schooltool.group.group import Group, GroupContainer

        >>> defineChecker(Group,
        ...               Checker({'title': 'zope.Public'},
        ...                       {'title': 'zope.Public'}))

    We want to display the generic groups a person is part of that aren't
    sections so we have a filter in the view:

        >>> groups = GroupContainer()
        >>> groups['tenth'] = tenth_grade = Group(title="Tenth Grade")
        >>> tenth_grade.members.add(student)
        >>> groups['team'] = team = Group(title="Sports Team")
        >>> team.members.add(student)

        >>> student_view = GroupsViewlet(student, TestRequest(), None, None)
        >>> student_view.update()
        >>> for schoolyear_data in student_view.schoolyears:
        ...     for group in schoolyear_data['groups']:
        ...         print group.title
        Sports Team
        Tenth Grade

    """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=doctest.ELLIPSIS|
                                                   doctest.REPORT_NDIFF))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Language independent date formatting.
"""
from zope.publisher.browser import BrowserView

from zope.i18n.locales import Locale
from zope.i18n.locales import LocaleIdentity


class LocaleLookupMixin(object):
    """Mixin for extracting of locale from request.

    Defaults to english if no locale is set in the request.
    """

    def getLocale(self):
        if not hasattr(self, 'request'):
            raise NotImplementedError("LocaleLookupMixin need to be applied on a view")

        if self.request.locale.id.language is None:
            # if we dont have any locale defined in the request
            # we set default to english
            id = LocaleIdentity('en', territory='US')
            locale = Locale(id)
        else:
            if hasattr(self.request, 'locale'):
                locale = self.request.locale

        return locale


class DateFormatterFullView( BrowserView, LocaleLookupMixin):
    """Formats the date using the 'full' format"""

    def __call__(self):
        locale = self.getLocale()
        formatter = locale.dates.getFormatter('date','full')
        return formatter.format(self.context)


class DateFormatterLongView( BrowserView, LocaleLookupMixin):
    """Formats the date using the 'long' format"""

    def __call__(self):
        locale = self.getLocale()
        formatter = locale.dates.getFormatter('date','long')
        return formatter.format(self.context)


class DateFormatterMediumView( BrowserView, LocaleLookupMixin):
    """Formats the date using the 'medium' format"""

    def __call__(self):
        locale = self.getLocale()
        formatter = locale.dates.getFormatter('date','medium')
        return formatter.format(self.context)


class DateFormatterShortView( BrowserView, LocaleLookupMixin):
    """Formats the date using the 'short' format"""

    def __call__(self):
        return self.context.strftime("%Y-%m-%d")



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional selenium tests for schooltool.schoolyear
"""
import unittest

from schooltool.testing.util import format_table
from schooltool.testing.selenium import collect_ftests
from schooltool.schoolyear.stesting import schoolyear_selenium_layer


def print_timetables_table(browser):
    driver = browser.driver
    for tr in driver.execute_script(
            'return $(arguments[0])', 'table tbody tr'):
        row = []
        for i, td in enumerate(driver.execute_script(
                'return $(arguments[0]).find("td")', tr)):
            e = td
            if i == 0:
                e = driver.execute_script(
                    'return $(arguments[0]).find("a")', td)[0]
            row.append(e.text)
        print ', '.join(row)


def print_timetable(browser):
    driver = browser.driver
    h3s = driver.execute_script('return $(arguments[0])', 'h3')
    tables = []
    for table in driver.execute_script('return $(arguments[0])', 'table'):
        rows = []
        row = []
        for th in driver.execute_script(
                'return $(arguments[0]).find(arguments[1])', table, 'thead th'):
            row.append(th.text)
        rows.append(row)
        for tr in driver.execute_script(
                'return $(arguments[0]).find(arguments[1])', table, 'tbody tr'):
            row = []
            for td in driver.execute_script(
                'return $(arguments[0]).find(arguments[1])', tr, 'td'):
                spans = []
                for span in driver.execute_script(
                        'return $(arguments[0]).find(arguments[1])',
                        td, 'span'):
                    text = span.text.strip()
                    if text:
                        spans.append(text)
                row.append(' '.join(spans))
            rows.append(row)
        tables.append(rows)
    print h3s[0].text
    print '-' * len(h3s[0].text)
    print h3s[1].text
    print format_table(tables[0], header_rows=1)
    print h3s[2].text
    print format_table(tables[1], header_rows=1)


def test_suite():
    extra_globs = {
        'print_timetables_table': print_timetables_table,
        'print_timetable': print_timetable,
        }
    return collect_ftests(layer=schoolyear_selenium_layer,
                          extra_globs=extra_globs)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Functional testing utilites for Contacts.
"""
from schooltool.testing.functional import TestBrowser


def addContact(firstname, lastname, address='', email='', browser=None):

    if browser is None:
        browser = TestBrowser('manager', 'schooltool')

    browser.getLink('Manage').click()
    browser.getLink('Contacts').click()
    browser.getLink('New Contact').click()

    browser.getControl('First name').value = firstname
    browser.getControl('Last name').value = lastname
    browser.getControl('Address line 1').value = address
    browser.getControl('Email').value = email

    browser.getControl('Add').click()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
course browser views.
"""

from urllib import urlencode

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.interface import implements
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.component import getUtility
from zope.container.interfaces import INameChooser
from zope.interface import directlyProvides
from zope.intid.interfaces import IIntIds
from zope.app.form.browser.add import AddView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import BrowserView
from zope.security.checker import canAccess
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.i18n.interfaces.locales import ICollator
from zope.i18n import translate
from zope.viewlet.viewlet import ViewletBase
from zc.table.interfaces import ISortableColumn
from z3c.form import field, button, form
from z3c.form.interfaces import HIDDEN_MODE

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.app import ContentTitle
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.common.inlinept import InheritTemplate
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.term.interfaces import ITerm
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.skin.containers import ContainerView
from schooltool.course.interfaces import ICourse, ICourseContainer
from schooltool.course.interfaces import ISectionContainer
from schooltool.course.interfaces import ILearner, IInstructor
from schooltool.course.course import Course
from schooltool.skin import flourish
from schooltool.skin.flourish.viewlet import Viewlet
from schooltool.skin.flourish.containers import ContainerDeleteView
from schooltool.skin.flourish.page import RefineLinksViewlet
from schooltool.skin.flourish.page import LinkViewlet
from schooltool.skin.flourish.page import Page
from schooltool.skin.flourish.page import Content
from schooltool.skin.flourish.page import ModalFormLinkViewlet
from schooltool.skin.flourish.form import Form
from schooltool.skin.flourish.form import AddForm
from schooltool.skin.flourish.form import DialogForm
from schooltool.skin.flourish.form import DisplayForm
from schooltool.skin.flourish.page import TertiaryNavigationManager
from schooltool import table
from schooltool.common import SchoolToolMessage as _


class CourseContainerAbsoluteURLAdapter(BrowserView):

    adapts(ICourseContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        container_id = int(self.context.__name__)
        int_ids = getUtility(IIntIds)
        container = int_ids.getObject(container_id)
        url = str(getMultiAdapter((container, self.request), name='absolute_url'))
        return url + '/courses'

    __call__ = __str__


class CourseContainerView(ContainerView):
    """A Course Container view."""

    __used_for__ = ICourseContainer

    index_title = _("Course index")

    @property
    def school_year(self):
        return ISchoolYear(self.context)


class CourseView(BrowserView):
    """A view for courses."""

    __used_for__ = ICourse

    @property
    def school_year(self):
        return ISchoolYear(self.context)

    @property
    def details(self):
        details = []
        for name in ['course_id', 'government_id', 'credits']:
            value = getattr(self.context, name)
            if value is not None and unicode(value).strip():
                details.append({
                    'title': ICourse[name].title,
                    'value': value,
                    })
        return details

    @property
    def sections(self):
        items = []
        for section in self.context.sections:
            term = ITerm(section, None)
            items.append({
                'section': section,
                'term': term,
                })
        def sortKey(item):
            # I consider it acceptable to violate security for sorting purposes
            # in this case.
            section = removeSecurityProxy(item['section'])
            term = removeSecurityProxy(item['term'])
            return u'%s%s%s' % (section.label,
                                term.first,
                                section.__name__)
        return sorted(items, key=sortKey)


class CourseAddView(AddView):
    """A view for adding Courses."""

    def nextURL(self):
        return absoluteURL(self.context.context, self.request)

    def update(self):
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
        else:
            return AddView.update(self)


class CoursesViewlet(ViewletBase):
    """A viewlet showing the courses a person is in."""

    def __init__(self, *args, **kw):
        super(CoursesViewlet, self).__init__(*args, **kw)
        self.collator = ICollator(self.request.locale)

    def update(self):
        self.instructorOf = self.sectionsAsTeacher()
        self.learnerOf = self.sectionsAsLearner()

    def isTeacher(self):
        """Find out if the person is an instructor for any sections."""
        return bool(self.instructorOf)

    def isLearner(self):
        """Find out if the person is a member of any sections."""
        return bool(self.learnerOf)

    def sectionsAsTeacher(self):
        """Get the sections the person instructs."""
        return self.sectionsAs(IInstructor)

    def sectionsAsLearner(self):
        """Get the sections the person is a member of."""
        return self.sectionsAs(ILearner)

    def sectionsAs(self, role_interface):
        schoolyears_data = {}
        for section in role_interface(self.context).sections():
            sy = ISchoolYear(section)
            if sy not in schoolyears_data:
                schoolyears_data[sy] = {}
            term = ITerm(section)
            if term not in schoolyears_data[sy]:
                schoolyears_data[sy][term] = []
            schoolyears_data[sy][term].append(section)
        result = []
        for sy in sorted(schoolyears_data, key=lambda x:x.first, reverse=True):
            sy_info = {'obj': sy, 'terms': []}
            for term in sorted(schoolyears_data[sy], key=lambda x:x.first):
                sortingKey = lambda section:{'course':
                                             ', '.join([course.title
                                                        for course in
                                                        section.courses]),
                                             'section_title': section.title}
                term_info = {'obj': term, 'sections': []}
                for section in sorted(schoolyears_data[sy][term],
                                      cmp=self.sortByCourseAndSection,
                                      key=sortingKey):
                    section_info = {'obj': section,
                                    'title': '%s -- %s' % \
                                    (', '.join(course.title
                                               for course in section.courses),
                                     section.title)}
                    term_info['sections'].append(section_info)
                sy_info['terms'].append(term_info)
            result.append(sy_info)
        return result

    def sortByCourseAndSection(self, this, other):
        if this['course'] is other['course']:
            return self.collator.cmp(this['section_title'],
                                     other['section_title'])
        return self.collator.cmp(this['course'], other['course'])


class FlourishCoursesViewlet(Viewlet):
    """A flourish viewlet showing the courses a person is in."""

    template = ViewPageTemplateFile('templates/f_coursesviewlet.pt')
    render = lambda self, *a, **kw: self.template(*a, **kw)

    def __init__(self, *args, **kw):
        super(FlourishCoursesViewlet, self).__init__(*args, **kw)
        self.collator = ICollator(self.request.locale)

    def update(self):
        self.instructorOf = self.sectionsAsTeacher()
        self.learnerOf = self.sectionsAsLearner()

    def isTeacher(self):
        """Find out if the person is an instructor for any sections."""
        return bool(self.instructorOf)

    def isLearner(self):
        """Find out if the person is a member of any sections."""
        return bool(self.learnerOf)

    def sectionsAsTeacher(self):
        """Get the sections the person instructs."""
        return self.sectionsAs(IInstructor)

    def sectionsAsLearner(self):
        """Get the sections the person is a member of."""
        return self.sectionsAs(ILearner)

    def sectionsAs(self, role_interface):
        schoolyears_data = {}
        for section in role_interface(self.context).sections():
            section = removeSecurityProxy(section)
            sy = ISchoolYear(section)
            if sy not in schoolyears_data:
                schoolyears_data[sy] = {}
            term = ITerm(section)
            if term not in schoolyears_data[sy]:
                schoolyears_data[sy][term] = []
            schoolyears_data[sy][term].append(section)
        result = []
        for sy in sorted(schoolyears_data, key=lambda x:x.first, reverse=True):
            sy_info = {'obj': sy, 'terms': []}
            for term in sorted(schoolyears_data[sy],
                               key=lambda x:x.first,
                               reverse=True):
                sortingKey = lambda section:{'course':
                                             ', '.join([course.title
                                                        for course in
                                                        section.courses]),
                                             'section_title': section.title}
                term_info = {'obj': term, 'sections': []}
                for section in sorted(schoolyears_data[sy][term],
                                      cmp=self.sortByCourseAndSection,
                                      key=sortingKey):
                    section_info = {
                        'obj': section,
                        'title': section.title,
                        }
                    term_info['sections'].append(section_info)
                sy_info['terms'].append(term_info)
            result.append(sy_info)
        return result

    def sortByCourseAndSection(self, this, other):
        if this['course'] is other['course']:
            return self.collator.cmp(this['section_title'],
                                     other['section_title'])
        return self.collator.cmp(this['course'], other['course'])


class CoursesTertiaryNavigationManager(TertiaryNavigationManager,
                                       ActiveSchoolYearContentMixin):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @property
    def items(self):
        result = []
        active = self.schoolyear
        schoolyears = active.__parent__ if active is not None else {}
        for schoolyear in schoolyears.values():
            url = '%s/%s?schoolyear_id=%s' % (
                absoluteURL(self.context, self.request),
                'courses',
                schoolyear.__name__)
            result.append({
                    'class': schoolyear.first == active.first and 'active' or None,
                    'viewlet': u'<a href="%s">%s</a>' % (url, schoolyear.title),
                    })
        return result


class CoursesActiveTabRefineLinksViewlet(RefineLinksViewlet,
                                         ActiveSchoolYearContentMixin):

    def __init__(self, context, request, view, manager):
        courses = self.getCourseContainer(context, request)
        RefineLinksViewlet.__init__(self, courses, request, view, manager)

    @classmethod
    def getCourseContainer(cls, context, request):
        courses = ICourseContainer(cls.schoolyear, context)
        return courses


class CoursesAddLinks(CoursesActiveTabRefineLinksViewlet):
    """Manager for Add links in CoursesView"""


class CoursesImportLinks(CoursesActiveTabRefineLinksViewlet):
    """Course import links viewlet."""


class CourseAddLinks(CoursesActiveTabRefineLinksViewlet):
    """Manager for Add links in CourseView"""

    def render(self):
        # This check is necessary because the user can be a leader
        # of the context course, which gives him schooltool.edit on it
        if canAccess(self.context.__parent__, '__delitem__'):
            return super(CourseAddLinks, self).render()


class CourseActionsLinks(RefineLinksViewlet):
    """Manager for Action links in CourseView"""

    def render(self):
        # This check is necessary because the user can be a leader
        # of the context course, which gives him schooltool.edit on it
        if canAccess(self.context.__parent__, '__delitem__'):
            return super(CourseActionsLinks, self).render()


class CourseDeleteLink(ModalFormLinkViewlet):

    @property
    def enabled(self):
        if not flourish.canDelete(self.context):
            return False
        return super(CourseDeleteLink, self).enabled

    @property
    def dialog_title(self):
        title = _(u'Delete ${course}',
                  mapping={'course': self.context.title})
        return translate(title, context=self.request)


class CourseAddLinkViewlet(LinkViewlet, ActiveSchoolYearContentMixin):

    @property
    def url(self):
        courses = ICourseContainer(self.schoolyear)
        return '%s/%s' % (absoluteURL(courses, self.request),
                          'addSchoolToolCourse.html')


class CourseAddLinkFromCourseViewlet(CourseAddLinkViewlet):

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @property
    def url(self):
        courses = ICourseContainer(self.schoolyear)
        return '%s/%s?camefrom=%s' % (
            absoluteURL(courses, self.request),
            'addSchoolToolCourse.html',
            absoluteURL(self.context, self.request))


class FlourishCoursesView(table.table.TableContainerView,
                          ActiveSchoolYearContentMixin):

    content_template = InlineViewPageTemplate('''
      <div>
        <tal:block content="structure view/container/schooltool:content/ajax/table" />
      </div>
    ''')

    @property
    def title(self):
        schoolyear = self.schoolyear
        return _('Courses for ${schoolyear}',
                 mapping={'schoolyear': schoolyear.title})

    @property
    def container(self):
        schoolyear = self.schoolyear
        return ICourseContainer(schoolyear)


def getCoursesTable(context, request, view, manager):
    container = view.container
    table = queryMultiAdapter(
        (container, request, view, manager),
        flourish.interfaces.IViewlet,
        'table')
    return table


class CoursesTableBase(table.ajax.Table):

    def columns(self):
        title = table.table.LocaleAwareGetterColumn(
            name='title',
            title=_(u'Title'),
            getter=lambda i, f: i.title,
            subsort=True)
        course_id = table.table.LocaleAwareGetterColumn(
            name='course_id',
            title=_('Course ID'),
            getter=lambda i, f: i.course_id or '',
            subsort=True)
        directlyProvides(title, ISortableColumn)
        directlyProvides(course_id, ISortableColumn)
        return [title, course_id]


class CoursesTable(CoursesTableBase):

    pass


class CourseListTable(CoursesTableBase):

    @property
    def source(self):
        schoolyear = ISchoolYear(self.context)
        return ICourseContainer(schoolyear)

    def items(self):
        return self.context.courses


class CourseTableSchoolYear(flourish.viewlet.Viewlet):
    template = InlineViewPageTemplate('''
      <input type="hidden" name="schoolyear_id"
             tal:define="schoolyear_id view/view/schoolyear/__name__|nothing"
             tal:condition="schoolyear_id"
             tal:attributes="value schoolyear_id" />
    ''')


class CourseTableDoneLink(flourish.viewlet.Viewlet):
    template = InlineViewPageTemplate('''
      <h3 tal:define="can_manage context/schooltool:app/schooltool:can_edit"
          class="done-link" i18n:domain="schooltool">
        <tal:block condition="can_manage">
          <a tal:attributes="href string:${context/schooltool:app/@@absolute_url}/manage"
             i18n:translate="">Done</a>
        </tal:block>
        <tal:block condition="not:can_manage">
          <a tal:attributes="href request/principal/schooltool:person/@@absolute_url"
             i18n:translate="">Done</a>
        </tal:block>
      </h3>
      ''')


class CourseContainerTitle(ContentTitle):

    @property
    def title(self):
        schoolyear = ISchoolYear(self.context)
        return _('Courses for ${schoolyear}',
                 mapping={'schoolyear': schoolyear.title})


class FlourishCourseContainerDeleteView(ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            schoolyear = ISchoolYear(self.context)
            params = {'schoolyear_id': schoolyear.__name__.encode('utf-8')}
            url = '%s/courses?%s' % (
                absoluteURL(ISchoolToolApplication(None), self.request),
                urlencode(params))
            return url
        return ContainerDeleteView.nextURL(self)


class FlourishCourseView(DisplayForm):

    template = InheritTemplate(Page.template)
    content_template = ViewPageTemplateFile('templates/f_course_view.pt')
    fields = field.Fields(ICourse)
    fields = fields.select('__name__', 'title', 'description', 'course_id', 'government_id', 'credits')

    @property
    def sections(self):
        return list(removeSecurityProxy(self.context).sections)

    @property
    def unproxied_course(self):
        return removeSecurityProxy(self.context)

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def title(self):
        schoolyear = ISchoolYear(self.context.__parent__)
        return _('Courses for ${schoolyear}',
                 mapping={'schoolyear': schoolyear.title})

    @property
    def subtitle(self):
        return self.context.title

    def done_link(self):
        schoolyear = ISchoolYear(self.context.__parent__)
        url = '%s/%s?schoolyear_id=%s' % (
            absoluteURL(ISchoolToolApplication(None), self.request),
            'courses',
            schoolyear.__name__)
        return url

    def updateWidgets(self):
        super(FlourishCourseView, self).updateWidgets()
        for widget in self.widgets.values():
            if not widget.value:
                widget.mode = HIDDEN_MODE

    def has_leaders(self):
        return bool(list(self.context.leaders))


class FlourishCourseViewDoneLink(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href view/view/done_link"
         i18n:translate="">Done</a>
    </h3>
    ''')


class FlourishCourseAddView(AddForm, ActiveSchoolYearContentMixin):

    template = InheritTemplate(Page.template)
    label = None
    legend = _('Course Information')
    fields = field.Fields(ICourse)
    fields = fields.select('title', 'description', 'course_id', 'government_id', 'credits')

    def updateActions(self):
        super(FlourishCourseAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(FlourishCourseAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        if 'camefrom' in self.request:
            url = self.request['camefrom']
            self.request.response.redirect(url)
            return
        app = ISchoolToolApplication(None)
        url = self.url_with_schoolyear_id(app, view_name='courses')
        self.request.response.redirect(url)

    def create(self, data):
        course = Course(data['title'], data.get('description'))
        form.applyChanges(self, course, data)
        return course

    def add(self, course):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', course)
        self.context[name] = course
        self._course = course
        return course

    def nextURL(self):
        return absoluteURL(self._course, self.request)

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @property
    def title(self):
        return _('Courses for ${schoolyear}',
                 mapping={'schoolyear': self.schoolyear.title})


class FlourishCourseEditView(Form, form.EditForm):

    template = InheritTemplate(Page.template)
    label = None
    legend = _('Course Information')
    fields = field.Fields(ICourse)
    fields = fields.select('title', 'description', 'course_id', 'government_id', 'credits')

    @property
    def title(self):
        return self.context.title

    def update(self):
        return form.EditForm.update(self)

    def updateActions(self):
        super(FlourishCourseEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(FlourishCourseEditView, self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class FlourishCourseDeleteView(DialogForm, form.EditForm):
    """View used for confirming deletion of a course."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishCourseDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishCourseFilterWidget(table.table.FilterWidget):

    template = ViewPageTemplateFile('templates/f_course_filter.pt')

    def filter(self, results):
        if 'SEARCH' in self.request:
            searchstr = self.request['SEARCH'].lower()
            results = [item for item in results
                       if searchstr in item.title.lower() or
                       (item.course_id and searchstr in item.course_id.lower())]
        return results


class CoursesTableFilter(table.ajax.TableFilter, FlourishCourseFilterWidget):

    title = _("Title or course ID")

    def filter(self, results):
        if self.ignoreRequest:
            return results
        return FlourishCourseFilterWidget.filter(self, results)


class FlourishCourseTableFormatter(table.table.SchoolToolTableFormatter):

    def columns(self):
        title = table.table.LocaleAwareGetterColumn(
            name='title',
            title=_(u'Title'),
            getter=lambda i, f: i.title,
            subsort=True)
        course_id = table.table.LocaleAwareGetterColumn(
            name='course_id',
            title=_('Course ID'),
            getter=lambda i, f: i.course_id or '',
            subsort=True)
        directlyProvides(title, ISortableColumn)
        directlyProvides(course_id, ISortableColumn)
        return [title, course_id]


class FlourishManageCoursesOverview(Content, ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_courses_overview.pt')

    @property
    def courses(self):
        return ICourseContainer(self.schoolyear, None)

    @property
    def sections(self):
        if self.has_schoolyear:
            result = []
            for term in self.schoolyear.values():
                sections = ISectionContainer(term)
                result.extend(list(sections.values()))
            return result

    @property
    def render_sections_link(self):
        return self.schoolyear is not None and \
               self.schoolyear and \
               self.courses is not None and \
               self.courses

    def courses_url(self):
        return self.url_with_schoolyear_id(self.context, view_name='courses')

    def sections_url(self):
        return self.url_with_schoolyear_id(self.context, view_name='sections')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Timetable builders.
"""
import itertools
import pytz
import datetime
from persistent.list import PersistentList

from zope.container.interfaces import INameChooser
from zope.component import getUtility
from zope.intid.interfaces import IIntIds

from schooltool.app.interfaces import IApplicationPreferences
from schooltool.generations.evolve36.helper import assert_not_broken
from schooltool.generations.evolve36.helper import BuildContext
from schooltool.timetable.app import SchoolToolSchedules
from schooltool.timetable.timetable import TimetableContainer
from schooltool.timetable.timetable import Timetable
from schooltool.timetable.daytemplates import WeekDayTemplates
from schooltool.timetable.daytemplates import SchoolDayTemplates
from schooltool.timetable.daytemplates import DayTemplate
from schooltool.timetable.daytemplates import TimeSlot
from schooltool.timetable.schedule import Period
from schooltool.timetable.schedule import MeetingException

APP_TIMETABLES_KEY = 'schooltool.timetable.timetables'


def createDayTemplates(timetable, factory, attr):
    schedule = factory()
    setattr(timetable, attr, schedule)
    schedule.__parent__ = timetable
    schedule.__name__ = unicode(attr)
    schedule.initTemplates()
    return schedule


class PeriodsBuilder(object):
    days = None

    def readPeriods(self, schema_day):
        periods = []
        for period_name in schema_day.periods:
            homeroom = bool(period_name in schema_day.homeroom_period_ids)
            periods.append({
                    'title': unicode(period_name),
                    'activity_type': homeroom and 'homeroom' or 'lesson',
                    })
        return periods

    def read(self, schema, context):
        assert_not_broken(schema)
        self.days = []
        for day_id, day in schema.items():
            assert_not_broken(day)
            self.days.append({
                    'id': day_id,
                    'title': unicode(day_id),
                    'periods': self.readPeriods(day),
                    })

    def addDayTemplate(self, templates, day_key, day):
        period_map = {}
        template = DayTemplate(title=day['title'])
        templates[day_key] = template

        name_chooser = INameChooser(template)
        for item in day['periods']:
            period = Period(title=item['title'],
                            activity_type=item['activity_type'])
            key = name_chooser.chooseName('', period)
            template[key] = period
            period_map[(day['id'], item['title'])] = period
        return BuildContext(period_map=period_map)


class WeekDayPeriodsBuilder(PeriodsBuilder):
    weekday_keys = ("Monday", "Tuesday", "Wednesday",
                    "Thursday", "Friday", "Saturday", "Sunday")
    timetable_day_keys = ()

    def read(self, schema, context):
        PeriodsBuilder.read(self, schema, context)
        if schema.model.timetableDayIds:
            self.timetable_day_keys = tuple(schema.model.timetableDayIds)

    def getDay(self, weekday):
        try:
            key = self.timetable_day_keys[weekday]
        except IndexError:
            key = self.weekday_keys[weekday]
        days = dict([(day['title'], day) for day in self.days])
        if key in days:
            return days[key]
        if 'None' in days:
            day = dict(days['None'])
            day['title'] = unicode(key) # XXX: hmmm
            return day
        return {'id': key,
                'title': unicode(key),
                'periods': []}

    def build(self, timetable, context):
        schedule = createDayTemplates(
            timetable, WeekDayTemplates, 'periods')

        result = BuildContext(period_map={})
        for weekday in range(7):
            day = self.getDay(weekday)
            key = unicode(weekday)
            built = self.addDayTemplate(schedule.templates, key, day)
            result.period_map.update(built.period_map)
        return result(schedule=schedule)


class SchoolDayPeriodsBuilder(PeriodsBuilder):

    def build(self, timetable, context):
        schedule = createDayTemplates(
            timetable, SchoolDayTemplates, 'periods')

        result = BuildContext(period_map={})
        for day in self.days:
            # XXX: title as key is not very safe, isn't it
            built = self.addDayTemplate(schedule.templates, day['title'], day)
            result.period_map.update(built.period_map)
        return result(schedule=schedule)


class TimeSlotsBuilder(object):
    schedule_factory = None
    schedule_attr = 'time_slots'
    days = None

    def readSlots(self, school_day_template):
        if school_day_template is None:
            return []

        assert_not_broken(school_day_template)
        assert_not_broken(*list(school_day_template))

        slots = [{'tstart': slot.tstart,
                  'duration': slot.duration,
                  'activity_type': None, # XXX: !!!
                  }
                 for slot in school_day_template]
        return slots

    def read(self, schema, context):
        raise NotImplementedError()

    def addDayTemplate(self, templates, day_key, day):
        template = DayTemplate(title=day['title'])
        templates[day_key] = template

        name_chooser = INameChooser(template)
        for item in day['time_slots']:
            time_slot = TimeSlot(
                item['tstart'], item['duration'],
                activity_type=item['activity_type'])
            key = name_chooser.chooseName('', time_slot)
            template[key] = time_slot


class WeekDayTimeSlotsBuilder(TimeSlotsBuilder):
    weekdays = ["Monday", "Tuesday", "Wednesday",
                "Thursday", "Friday", "Saturday", "Sunday"]

    def read(self, schema, context):
        model = schema.model
        assert_not_broken(model)

        self.days = []
        for weekday in range(7):
            template = model.dayTemplates.get(weekday)
            if template is None:
                template = model.dayTemplates.get(None)

            slots = self.readSlots(template)
            self.days.append({
                    'title': self.weekdays[weekday],
                    'time_slots': slots,
                    })

    def build(self, timetable, context):
        schedule = createDayTemplates(
            timetable, WeekDayTemplates, 'time_slots')

        for weekday, day in enumerate(self.days):
            key = unicode(weekday)
            self.addDayTemplate(schedule.templates, key, day)
        return BuildContext(schedule=schedule)


class SchoolDayTimeSlotsBuilder(TimeSlotsBuilder):
    def read(self, schema, context):
        model = schema.model
        assert_not_broken(model)

        self.days = []
        for day_id in model.timetableDayIds:
            template = model.dayTemplates[day_id]
            slots = self.readSlots(template)
            self.days.append({
                    'title': unicode(day_id),
                    'time_slots': slots,
                    })

    def build(self, timetable, context):
        schedule = createDayTemplates(
            timetable, SchoolDayTemplates, 'time_slots')

        for day in self.days:
            # XXX: title as key is not very safe, isn't it
            self.addDayTemplate(schedule.templates, day['title'], day)
        return BuildContext(schedule=schedule)


class ExceptionDayBuilder(object):

    def getOriginalDayId(self, term, schema, day_date):
        raise NotImplementedError()

    def getTimeSlotTemplate(self, schema, date, day_id):
        raise NotImplementedError()

    def readReplacementDay(self, schema, date):
        day_id = schema.model.exceptionDayIds[date]
        schema_day = schema[day_id]
        time_slots = self.getTimeSlotTemplate(schema, date, day_id)

        result = [
            (date, day_id, period, slot.tstart, slot.duration)
            for (period, slot) in zip(schema_day.keys()), sorted(time_slots)]
        return result

    def readExceptionDay(self, schema, date, day_id):
        exception_day = schema.model.exceptionDays[date]
        schema_day = schema[day_id]
        schema_periods = schema_day.keys()
        result = [
            (date, day_id, period, slot.tstart, slot.duration)
            for (period, slot) in exception_day
            if period in schema_periods]
        return result

    def getTerm(self, schoolyear, date):
        for term in schoolyear.values():
            if date in term:
                return term
        return None

    def read(self, schema, context):
        self.templates = []

        for date in schema.model.exceptionDays:
            term = self.getTerm(context.schoolyear, date)
            if term is None:
                continue
            if date in schema.model.exceptionDayIds:
                day_id = schema.model.exceptionDayIds[date]
            else:
                day_id = self.getOriginalDayId(term, schema, date)
            self.templates.append(self.readExceptionDay(schema, date, day_id))

        for date in schema.model.exceptionDayIds:
            if date not in schema.model.exceptionDays:
                self.templates.append(self.readReplacementDay(schema, date))

    def build(self, timetable, context):
        period_map = context.period_map
        tz = pytz.timezone(timetable.timezone)
        by_date = {}

        for day_templates in self.templates:
            for n, info in enumerate(day_templates):
                date, day_id, period_id, tstart, duration = info

                period_key = (day_id, period_id)
                period = period_map.get(period_key)

                dtstart = datetime.datetime.combine(date, tstart)
                dtstart = dtstart.replace(tzinfo=tz)
                if period is None:
                    meeting_id = None
                else:
                    meeting_id = timetable.periodMeetingId(
                        date, period, n+1)
                meeting = MeetingException(
                    dtstart, duration,
                    period=period,
                    meeting_id=meeting_id)
                if date not in by_date:
                    by_date[date] = []
                by_date[date].append(meeting)

        for date in by_date:
            timetable.exceptions[date] = PersistentList(
                sorted(by_date[date], key=lambda m: m.dtstart))
        return BuildContext(exceptions=timetable.exceptions)


class SequentialModelExceptions(ExceptionDayBuilder):

    def getOriginalDayId(self, term, schema, day_date):
        generator = itertools.cycle(schema.model.timetableDayIds)
        for date in term:
            if date == day_date:
                break
            if term.isSchoolday(date):
                if date not in schema.model.exceptionDayIds:
                    generator.next()
        if not term.isSchoolday(day_date):
            return None
        return generator.next()


class SequentialDaysModelExceptions(SequentialModelExceptions):

    def getTimeSlotTemplate(self, schema, date, day_id):
        return self.dayTemplates[date.weekday()]


class SequentialDayIdModelExceptions(SequentialModelExceptions):

    def getTimeSlotTemplate(self, schema, date, day_id):
        return schema.dayTemplates[day_id]


class WeeklyModelExceptions(ExceptionDayBuilder):

    def getOriginalDayId(self, term, schema, day_date):
        try:
            return schema.model.timetableDayIds[day_date.weekday()]
        except IndexError:
            return None

    def getTimeSlotTemplate(self, schema, date, day_id):
        return self.dayTemplates[date.weekday()]


class TimetableBuilder(object):

    __name__ = None
    timezone = None
    title = None

    periods = None
    time_slosts = None
    exceptions = None
    set_default = False

    def read(self, schema, context):
        assert_not_broken(schema)

        self.timezone = unicode(schema.timezone)
        self.title = unicode(schema.title)
        self.__name__ = schema.__name__

        model_name = schema.model.__class__.__name__

        if model_name == 'SequentialDaysTimetableModel':
            self.periods = SchoolDayPeriodsBuilder()
            self.time_slots = WeekDayTimeSlotsBuilder()
            self.exceptions = SequentialDaysModelExceptions()

        elif model_name == 'SequentialDayIdBasedTimetableModel':
            self.periods = SchoolDayPeriodsBuilder()
            self.time_slots = SchoolDayTimeSlotsBuilder()
            self.exceptions = SequentialDayIdModelExceptions()

        elif model_name == 'WeeklyTimetableModel':
            self.periods = WeekDayPeriodsBuilder()
            self.time_slots = WeekDayTimeSlotsBuilder()
            self.exceptions = WeeklyModelExceptions()

        self.periods.read(schema, context())
        self.time_slots.read(schema, context())
        self.exceptions.read(schema, context())

    def build(self, timetables, context):
        # XXX: what if timezone is broken or outdated?
        if self.timezone is None:
            timezone = IApplicationPreferences(context.app).timezone
        else:
            timezone = self.timezone

        first, last = context.schoolyear.first, context.schoolyear.last

        timetable = Timetable(
            first, last,
            title=self.title,
            timezone=timezone)
        timetables[self.__name__] = timetable
        timetable.__parent__ = timetables

        built_periods = self.periods.build(
            timetable, context(timetables=timetables))
        built_time_slots = self.time_slots.build(
            timetable, context(timetables=timetables))
        built_exceptions = self.exceptions.build(
            timetable, context(timetables=timetables,
                               period_map=built_periods.period_map))

        if self.set_default:
            timetables.default = timetable

        return BuildContext(timetable=timetable,
                            schema_id=self.__name__,
                            period_map=built_periods.period_map)


class TimetableContainerBuilder(object):
    year_int_id = None
    timetables = None

    def read(self, schema_container, context):
        assert_not_broken(schema_container)

        self.year_int_id = context.year_int_id
        schoolyear = getUtility(IIntIds).getObject(self.year_int_id)

        default_id = schema_container._default_id

        self.timetables = []
        for key, schema in list(schema_container.items()):
            assert_not_broken(schema)
            builder = TimetableBuilder()
            if (default_id is not None and default_id == key):
                builder.set_default = True
            builder.read(schema, context(schema_container=schema_container,
                                         schoolyear=schoolyear))
            self.timetables.append(builder)

    def build(self, timetable_root, context):
        key = unicode(self.year_int_id)
        container = timetable_root[key] = TimetableContainer()
        schoolyear = getUtility(IIntIds).getObject(self.year_int_id)

        result = BuildContext(period_map={})

        for builder in self.timetables:
            built = builder.build(
                container, context(timetable_root=timetable_root,
                                   schoolyear=schoolyear))

            schema_period_map = dict(
                [((built.schema_id, day_id, period_id), period)
                 for (day_id, period_id), period
                 in built.period_map.items()])

            result.period_map.update(schema_period_map)

        return result(timetables=container,
                      year_int_id=self.year_int_id)


class SchoolTimetablesBuilder(object):
    builders = None

    def read(self, app, context):
        assert_not_broken(app)
        schema_root = app['schooltool.timetable.schooltt']
        self.builders = []
        for container in schema_root.values():
            year_int_id = int(container.__name__)
            schoolyear = getUtility(IIntIds).queryObject(year_int_id)
            if schoolyear is None:
                # Dirty database: year was deleted, but timetable schemas
                #                 were left
                continue

            builder = TimetableContainerBuilder()

            assert_not_broken(container)
            builder.read(container, context(app=app,
                                            year_int_id=year_int_id))
            self.builders.append(builder)

    def clean(self, app, context):
        del app['schooltool.timetable.schooltt']

    def build(self, app, context):
        if APP_TIMETABLES_KEY not in app:
            app[APP_TIMETABLES_KEY] = SchoolToolSchedules()
        timetable_root = app[APP_TIMETABLES_KEY]

        result = BuildContext(period_map={})

        for builder in self.builders:
            built = builder.build(timetable_root, context(app=app))

            year_period_map = dict(
                [((built.year_int_id, schema_id, day_id, period_id), period)
                 for (schema_id, day_id, period_id), period
                 in built.period_map.items()])

            result.period_map.update(year_period_map)

        return result(timetable_root=timetable_root)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.calendar
"""

import unittest
import doctest


def doctest_interfaces():
    """Look for syntax errors in interfaces.py

        >>> import schooltool.calendar.interfaces

    """


def doctest_CalendarMixin_expand():
    """Tests for CalendarMixin.expand.

    Let's define a calendar that uses CalendarMixin and contains a fixed
    set of events

        >>> from datetime import datetime, timedelta
        >>> from schooltool.calendar.mixins import CalendarMixin
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule
        >>> Event = SimpleCalendarEvent # shorter

        >>> from pytz import utc

        >>> class MyCalendar(CalendarMixin):
        ...     def __iter__(self):
        ...         return iter([Event(datetime(2004, 12, 14, 12, 30,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'a'),
        ...                      Event(datetime(2004, 12, 15, 16, 30,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'c'),
        ...                      Event(datetime(2004, 12, 15, 14, 30,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'b'),
        ...                      Event(datetime(2004, 12, 16, 17, 30,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'd'),
        ...                      Event(datetime(2005,  2,  3,  4,  5,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=4), 'simple'),
        ...                      Event(datetime(2005,  2,  4,  4,  5,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=4), 'recurring',
        ...                            recurrence=DailyRecurrenceRule()),
        ...                      Event(datetime(2005, 6, 16, 23, 30,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'e'),
        ...                      Event(datetime(2005, 6, 17, 5, 30,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'f'),
        ...                      Event(datetime(2005, 6, 18, 1, 0,
        ...                                     tzinfo=utc),
        ...                            timedelta(hours=1), 'g'),
        ...                     ])

        >>> cal = MyCalendar()

    We will define a convenience function for showing all events returned
    by expand:

        >>> def show(first, last):
        ...     events = list(cal.expand(first, last))
        ...     events.sort()
        ...     print '[%s]' % ', '.join([e.title for e in events])

        >>> def show_long(first, last):
        ...     events = list(cal.expand(first, last))
        ...     events.sort()
        ...     for e in events:
        ...         print e.dtstart.strftime('%Y-%m-%d'), e.title

    Events that fall inside the interval

        >>> show(datetime(2004, 12, 1, tzinfo=utc),
        ...      datetime(2004, 12, 31, tzinfo=utc))
        [a, b, c, d]

        >>> show(datetime(2004, 12, 15, tzinfo=utc),
        ...      datetime(2004, 12, 16, tzinfo=utc))
        [b, c]

    Events that fall partially in the interval

        >>> show(datetime(2004, 12, 15, 17, 0, tzinfo=utc),
        ...      datetime(2004, 12, 16, 18, 0, tzinfo=utc))
        [c, d]

    Corner cases: if event.dtstart + event.duration == first, or
    event.dtstart == last, the event is not included.

        >>> show(datetime(2004, 12, 15, 15, 30, tzinfo=utc),
        ...      datetime(2004, 12, 15, 16, 30, tzinfo=utc))
        []

    Recurring events:

        >>> show_long(datetime(2005, 2, 2, tzinfo=utc),
        ...           datetime(2005, 2, 5, tzinfo=utc))
        2005-02-03 simple
        2005-02-04 recurring

        >>> show_long(datetime(2005, 2, 2, tzinfo=utc),
        ...           datetime(2005, 2, 6, tzinfo=utc))
        2005-02-03 simple
        2005-02-04 recurring
        2005-02-05 recurring

        >>> show_long(datetime(2005, 2, 10, tzinfo=utc),
        ...           datetime(2005, 2, 13, tzinfo=utc))
        2005-02-10 recurring
        2005-02-11 recurring
        2005-02-12 recurring

    Recurring events are replaced by proxy objects

        >>> from schooltool.calendar.interfaces import IExpandedCalendarEvent
        >>> events = list(cal.expand(datetime(2005, 2, 2, tzinfo=utc),
        ...                          datetime(2005, 2, 6, tzinfo=utc)))
        >>> events.sort()
        >>> [IExpandedCalendarEvent.providedBy(e) for e in events]
        [False, True, True]
        >>> events[1].original is events[2].original
        True

    When we expand with a different timezone, we see the events that occur on
    that date in the given timezone.

        >>> from pytz import timezone
        >>> eastern = timezone('US/Eastern')
        >>> show(datetime(2005, 6, 17, tzinfo=eastern),
        ...      datetime(2005, 6, 18, tzinfo=eastern))
        [recurring, f, g, recurring]

        >>> vilnius = timezone('Europe/Vilnius')
        >>> show(datetime(2005, 6, 17, tzinfo=vilnius),
        ...      datetime(2005, 6, 18, tzinfo=vilnius))
        [e, recurring, f]

    You can even mix timezones:

        >>> show(datetime(2005, 6, 17, tzinfo=utc),
        ...      datetime(2005, 6, 18, tzinfo=vilnius))
        [e, recurring, f]

        >>> show(datetime(2005, 6, 17, tzinfo=vilnius),
        ...      datetime(2005, 6, 18, tzinfo=utc))
        [e, recurring, f]

    """


def doctest_CalendarEventMixin_hasOccurrences():
    """Tests for CalendarEventMixin.hasOccurrences.

    We will use SimpleCalendarEvent which is a trivial subclass of
    CalendarEventMixin

        >>> from datetime import date, datetime, timedelta
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule

    A simple event always has occurrences.

        >>> e1 = SimpleCalendarEvent(datetime(2004, 11, 25, 12, 0),
        ...                          timedelta(minutes=10), 'whatever')
        >>> e1.hasOccurrences()
        True

    A forever-repeating event always has occurrences.

        >>> e2 = SimpleCalendarEvent(datetime(2004, 11, 25, 12, 0),
        ...                          timedelta(minutes=10), 'whatever',
        ...                          recurrence=DailyRecurrenceRule())
        >>> e2.hasOccurrences()
        True

    Here's an event without occurrences:

        >>> e3 = SimpleCalendarEvent(datetime(2004, 11, 25, 12, 0),
        ...                          timedelta(minutes=10), 'whatever',
        ...                          recurrence=DailyRecurrenceRule(
        ...                              count=3,
        ...                              exceptions=[date(2004, 11, 25),
        ...                                          date(2004, 11, 26),
        ...                                          date(2004, 11, 27)]))
        >>> e3.hasOccurrences()
        False

    However remove one exception, and it becomes an occurrence:

        >>> e4 = SimpleCalendarEvent(datetime(2004, 11, 25, 12, 0),
        ...                          timedelta(minutes=10), 'whatever',
        ...                          recurrence=DailyRecurrenceRule(
        ...                              count=3,
        ...                              exceptions=[date(2004, 11, 25),
        ...                                          date(2004, 11, 27)]))
        >>> e4.hasOccurrences()
        True

    """


def doctest_CalendarEventMixin_replace():
    """Make sure CalendarEventMixin.replace does not forget any attributes.

        >>> from schooltool.calendar.interfaces import ICalendarEvent
        >>> from zope.schema import getFieldNames
        >>> all_attrs = getFieldNames(ICalendarEvent)

    We will use SimpleCalendarEvent which is a trivial subclass of
    CalendarEventMixin

        >>> from datetime import datetime, timedelta
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> e1 = SimpleCalendarEvent(datetime(2004, 12, 15, 18, 57),
        ...                          timedelta(minutes=15),
        ...                          'Work on schooltool.calendar.simple')

        # XXX Need a better test for this.  event.dtstart should always be a
        # datetime object not a string
        #>>> for attr in all_attrs:
        #...     e2 = e1.replace(**{attr: 'new value'})
        #...     assert getattr(e2, attr) == 'new value', attr
        #...     assert e2 != e1, attr
        #...     assert e2.replace(**{attr: getattr(e1, attr)}) == e1, attr

    """


def doctest_CalendarEventMixin_expand():
    """Test expanding of recurring events

        >>> from schooltool.calendar.simple import SimpleCalendarEvent

    For non-recurring events, expand yields the event if it is in the
    range passed as arguments:

        >>> from pytz import utc
        >>> from datetime import datetime, timedelta
        >>> e1 = SimpleCalendarEvent(datetime(2004, 12, 15, 18, 57, tzinfo=utc),
        ...                          timedelta(minutes=15),
        ...                          'Work on schoolbell.calendar.simple')
        >>> expanded = list(e1.expand(
        ...     datetime(2004, 12, 15, 0, 0, tzinfo=utc),
        ...     datetime(2004, 12, 16, 0, 0, tzinfo=utc)))
        >>> expanded
        [<schooltool.calendar.simple.SimpleCalendarEvent object at ...>]
        >>> expanded[0] is e1
        True

    If the event is outside the datetime range passed, the method
    yields nothing.

        >>> list(e1.expand(datetime(2004, 12, 16, 0, 0, tzinfo=utc),
        ...                datetime(2004, 12, 17, 0, 0, tzinfo=utc)))
        []

    See doctest_CalendarMixin_expand doctest for an elaborate
    demonstration of the expanding functionality.
    """


def doctest_CalendarEventMixin_expand_at_midnight():
    """Regression tests for CalendarEventMixin.expand.

    Bug: an event that occurs at midnight and is 0 minutes long gets lost.

        >>> from pytz import utc
        >>> from datetime import datetime, timedelta
        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule
        >>> e1 = SimpleCalendarEvent(datetime(2005, 3, 2, 0, 0, tzinfo=utc),
        ...                          timedelta(0), "Corner case")
        >>> e2 = SimpleCalendarEvent(datetime(2005, 3, 4, 0, 0, tzinfo=utc),
        ...                          timedelta(0), "Recurring case",
        ...                          recurrence=DailyRecurrenceRule())

        >>> def show(first, last):
        ...     for event in e1, e2:
        ...         print [e.title for e in event.expand(first, last)]

        >>> show(datetime(2005, 3, 1, tzinfo=utc),
        ...      datetime(2005, 3, 2, tzinfo=utc))
        []
        []

        >>> show(datetime(2005, 3, 2, tzinfo=utc),
        ...      datetime(2005, 3, 3, tzinfo=utc))
        ['Corner case']
        []

        >>> show(datetime(2005, 3, 3, tzinfo=utc),
        ...      datetime(2005, 3, 4, tzinfo=utc))
        []
        []

        >>> show(datetime(2005, 3, 4, tzinfo=utc),
        ...      datetime(2005, 3, 5, tzinfo=utc))
        []
        ['Recurring case']

    """


def doctest_weeknum_bounds():
    """Unit test for schooltool.calendar.utils.weeknum_bounds.

    Check that weeknum_bounds is the reverse of datetime.isocalendar().

        >>> from datetime import date
        >>> from schooltool.calendar.utils import weeknum_bounds
        >>> d = date(2000, 1, 1)
        >>> while d < date(2010, 1, 1):
        ...     year, weeknum, weekday = d.isocalendar()
        ...     l, h = weeknum_bounds(year, weeknum)
        ...     assert l <= d <= h
        ...     d += d.resolution

    """


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(optionflags=doctest.ELLIPSIS))
    suite.addTest(doctest.DocFileSuite('../README.txt'))
    suite.addTest(doctest.DocTestSuite('schooltool.calendar.mixins'))
    suite.addTest(doctest.DocTestSuite('schooltool.calendar.simple'))
    suite.addTest(doctest.DocTestSuite('schooltool.calendar.recurrent'))
    suite.addTest(doctest.DocTestSuite('schooltool.calendar.utils'))
    suite.addTest(doctest.DocTestSuite('schooltool.calendar.browser',
                        optionflags=doctest.ELLIPSIS | doctest.REPORT_UDIFF))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool pluggable sample data generator

In order to be able to test SchoolTool with data sets comparable in
size to those used in real life routinely, we have this sample data
generation system.  Sample data is created by a variety of plugins
that can state their dependencies on other plugins explicitly.
Extension writers can provide plugins for sample data for their
extensions, building upon the data constructed by the plugins provided
with schooltool.  Dependencies must be acyclic.

Sample data generation plugins are named local utilities with an
interface schooltool.sampledata.interfaces.ISampleDataGeneratorPlugin.

$Id$
"""
import time

import transaction
from zope.component import getUtilitiesFor

from schooltool.sampledata.interfaces import ISampleDataPlugin
from schooltool.sampledata.interfaces import CyclicDependencyError


def generate(app, seed=None, dry_run=False, pluginNames=[]):
    """Generate sample data provided by all plugins.

    Runs the generate functions of all plugins in an order such that
    the dependencies are all generated before the dependents.

    In essence, this function performs a topological sort in a
    directed acyclic graph.

    Raises a CyclicDependencyError if the dependency graph has a cycle.

    Returns a dict with names of plugins run as keys and CPU times as
    values.
    """

    plugins = dict([(obj.name, obj) for name, obj in
                    getUtilitiesFor(ISampleDataPlugin)])


    # status is a dict plugin names as keys and statuses as values.
    # Statuses can be as follows:
    #
    #   new    -- not yet visited
    #   closed -- already processed
    #   open   -- being processed
    #
    # Stumbling over an 'open' node means there is a cyclic dependency

    new = 'new'
    open = 'open'
    closed = 'closed'
    status = dict([(name, new) for name in plugins])

    # dict with used CPU cycles for each plugin
    times = {}

    def visit(name):
        """The recursive part of the topological sort

        Raises a CyclicDependencyError if cyclic depencencies are found.
        """
        if status[name] == new:
            status[name] = open
            plugin = plugins[name]
            for dep in plugin.dependencies:
                visit(dep)
            start = time.clock()
            if not dry_run:
                plugin.generate(app, seed)
                # Some plugins can generate a lot of data, so we are
                # using savepoints to save on memory consuption.
                transaction.savepoint(optimistic=True)

            times[name] = time.clock() - start
            status[name] = closed

        elif status[name] == closed:
            return

        elif status[name] == open:
            raise CyclicDependencyError("cyclic dependency at '%s'" % name)


    for name in pluginNames:
        visit(name)

    return times



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 35.

Evolution script to unregister old catalog utilities.
"""
import zope.catalog
import zope.catalog.catalog
import zope.catalog.interfaces
import zope.intid
import zope.intid.interfaces
from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication
from zope.component import queryUtility, getUtility
from zope.component.hooks import getSite, setSite
from zope.traversing.api import traverse

from schooltool.testing.mock import ModulesSnapshot
from schooltool.app.interfaces import ISchoolToolApplication
from zope.catalog.interfaces import ICatalog


CATALOG_KEYS = [
    'schooltool.basicperson',
    'schooltool.contact',
    'schooltool.person',
    ]


def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)

    # Mock the renaming of zope.app.catalog and zope.app.intid to
    # zope.catalog and zope.intid.
    # This is for Critical Links deployments only - this part of Zope's evolution
    # was missed somehow.
    modules = ModulesSnapshot()
    modules.mock_module('zope.app.catalog')
    modules.mock_module('zope.app.catalog.interfaces')
    modules.mock_module('zope.app.catalog.catalog')
    modules.mock_module('zope.app.intid')
    modules.mock_module('zope.app.intid.interfaces')

    modules.mock_attr('zope.app.catalog.catalog', 'Catalog', zope.catalog.catalog.Catalog)
    modules.mock_attr('zope.app.catalog.interfaces', 'ICatalog', zope.catalog.interfaces.ICatalog)

    modules.mock_attr('zope.app.intid', 'IntIds', zope.intid.IntIds)
    modules.mock_attr('zope.app.intid.interfaces', 'IIntIds', zope.intid.interfaces.IIntIds)

    # Proceed with normal evolution now.

    old_site = getSite()
    apps = findObjectsProviding(root, ISchoolToolApplication)
    for app in apps:
        setSite(app)
        sm = app.getSiteManager()
        default = traverse(app, '++etc++site/default')

        intids = getUtility(zope.intid.interfaces.IIntIds)
        intids._p_changed = True

        for key in CATALOG_KEYS:
            util = queryUtility(ICatalog, name=key, default=None)
            if util is None:
                continue
            name = util.__name__
            sm.unregisterUtility(util, ICatalog, key)
            del default[name]

    setSite(old_site)

    modules.restore()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for groups
"""
import unittest
import doctest

from zope.interface import directlyProvides
from zope.interface.verify import verifyObject
from zope.app.testing import setup
from zope.traversing.interfaces import IContainmentRoot
from zope.interface import implements

from schooltool.testing.util import run_unit_tests
from schooltool.securitypolicy.interfaces import IAccessControlCustomisations
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IAsset


def doctest_GroupContainer():
    """Tests for GroupContainer

        >>> from schooltool.group.interfaces import IGroupContainer
        >>> from schooltool.group.group import GroupContainer
        >>> c = GroupContainer()
        >>> verifyObject(IGroupContainer, c)
        True

    Let's make sure it acts like a proper container should act

        >>> from zope.container.tests.test_btree import TestBTreeContainer
        >>> class Test(TestBTreeContainer):
        ...    def makeTestObject(self):
        ...        return GroupContainer()
        >>> run_unit_tests(Test)

    """


def doctest_Group():
    r"""Tests for Group

        >>> from schooltool.group.interfaces import IGroupContained
        >>> from schooltool.group.group import Group
        >>> group = Group()
        >>> verifyObject(IGroupContained, group)
        True

    Groups can have titles and descriptions too

        >>> illuminati = Group(title='Illuminati', description='Secret Group')
        >>> illuminati.title
        'Illuminati'
        >>> illuminati.description
        'Secret Group'

    """


def doctest_GroupInit():
    """Tests for GroupInit

    This test needs annotations, dependencies and traversal

        >>> setup.placelessSetUp()
        >>> setup.setUpAnnotations()
        >>> setup.setUpDependable()
        >>> setup.setUpTraversal()

        >>> from schooltool.group.group import GroupInit
        >>> from schooltool.app.app import SchoolToolApplication
        >>> app = SchoolToolApplication()

    we want app to have a location

        >>> directlyProvides(app, IContainmentRoot)

    When you call the initialization adapter

        >>> plugin = GroupInit(app)
        >>> plugin()

    it adds a container for groups

        >>> app['schooltool.group']
        <schooltool.group.group.GroupContainerContainer object at ...>

    Clean up

        >>> setup.placelessTearDown()

    """


def doctest_GroupCalendarViewersCrowd():
    """Tests for ConfigurableCrowd.

    Some setup:

        >>> setup.placelessSetUp()

        >>> setting = True
        >>> class CustomisationsStub(object):
        ...     implements(IAccessControlCustomisations)
        ...     def get(self, key):
        ...         print 'Getting %s' % key
        ...         return setting

        >>> class PersonsStub(object):
        ...     super_user = None

        >>> class AppStub(object):
        ...     implements(ISchoolToolApplication)
        ...     def __conform__(self, iface):
        ...         if iface == IAccessControlCustomisations:
        ...             return CustomisationsStub()
        ...     def __getitem__(self, name):
        ...         if name == 'persons':
        ...             return PersonsStub()

        >>> from zope.component import provideAdapter
        >>> provideAdapter(lambda context: AppStub(),
        ...                adapts=[None],
        ...                provides=ISchoolToolApplication)

        >>> member_of = ()
        >>> from schooltool.person.interfaces import IPerson
        >>> class PrincipalStub(object):
        ...     def __init__(self, name):
        ...         self.name = name
        ...     def __conform__(self, iface):
        ...         if iface == IPerson: return "IPerson(%s)" % self.name
        ...     class groups(object):
        ...         def __contains__(self, item):
        ...             print "Checking for membership in group %s" % item
        ...             return item in member_of
        ...     groups = groups()

        >>> from schooltool.group.group import GroupCalendarViewersCrowd

    Off we go:

        >>> group_val = True
        >>> leader_val = True
        >>> class GroupStub(object):
        ...     implements(IAsset)
        ...     class members(object):
        ...         def __contains__(self, item):
        ...             print "Checking for membership of %s in a group" % item
        ...             return group_val
        ...     members = members()
        ...     class leaders(object):
        ...         def __contains__(self, item):
        ...             print "Checking for leadership of %s in a group" % item
        ...             return leader_val
        ...     leaders = leaders()

        >>> crowd = GroupCalendarViewersCrowd(GroupStub())
        >>> crowd.contains("Principal")
        Getting everyone_can_view_group_calendar
        True

    If setting is set to False, we should still check for membership:

        >>> setting = False
        >>> crowd.contains(PrincipalStub("Principal"))
        Getting everyone_can_view_group_calendar
        Checking for membership of IPerson(Principal) in a group
        True

    If membership fails, we check leadership:

        >>> group_val = False
        >>> crowd.contains(PrincipalStub("Principal"))
        Getting everyone_can_view_group_calendar
        Checking for membership of IPerson(Principal) in a group
        Checking for leadership of IPerson(Principal) in a group
        True

    If leadership fails, check for membership in system groups

        >>> leader_val = False
        >>> crowd.contains(PrincipalStub("Principal"))
        Getting everyone_can_view_group_calendar
        Checking for membership of IPerson(Principal) in a group
        Checking for leadership of IPerson(Principal) in a group
        Checking for membership in group sb.group.administrators
        Checking for membership in group sb.group.manager
        Checking for membership in group sb.group.clerks
        False

    Add to administrators

        >>> member_of = ['sb.group.administrators']
        >>> crowd.contains(PrincipalStub("Principal"))
        Getting everyone_can_view_group_calendar
        Checking for membership of IPerson(Principal) in a group
        Checking for leadership of IPerson(Principal) in a group
        Checking for membership in group sb.group.administrators
        True

    Clean up

        >>> setup.placelessTearDown()

    """


def test_suite():
    return unittest.TestSuite([
                doctest.DocTestSuite(optionflags=doctest.ELLIPSIS),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Pluggable Traverser Interfaces

This implementation is independent of the presentation type. Sub-interfaces
must be written for every specific presentation type.
"""

from zope.interface import Interface, Attribute
from zope.publisher.interfaces import IPublishTraverse


class IPluggableTraverser(IPublishTraverse):
    """A pluggable traverser.

    This traverser traverses a name by utilizing helper traversers that are
    registered as ``ITraverserPlugin`` subscribers.
    """


class ISchoolToolTraverser(Interface):
    """A plugin for the pluggable traverser."""

    context = Attribute("The context object the plugin traverses")
    request = Attribute("The request object driving the traversal")

    def traverse(name):
        """The 'name' argument is the name that is to be looked up; it must
        be an ASCII string or Unicode object.

        If a lookup is not possible, raise a NotFound error.

        This method should return an object having the specified name and
        `self` as parent.
        """


class ITraverser(IPublishTraverse,
                 ISchoolToolTraverser):
    """A simple traverser.

    This traverser traverses a name by utilizing helper traversers that are
    registered as ``ITraverserPlugin`` subscribers.
    """


class ITraverserPlugin(ISchoolToolTraverser):
    """A plugin for the pluggable traverser."""




#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Level interfaces.
"""

import zope.schema
from zope.container.interfaces import IContainer, IContained
from zope.container.constraints import contains, containers
from zope.interface import Interface, Attribute

from schooltool.common import SchoolToolMessage as _


class ILevel(Interface):
    """A level of learing (basically Nth year of courses)."""

    title = zope.schema.TextLine(
        title=_("Title"))

    courses = Attribute(
        """Courses available for this level.
           see schooltool.relationship.interfaces.IRelationshipProperty.""")


class ILevelContainer(IContainer):
    """Container of levels."""

    contains(ILevel)


class ILevelContained(ILevel, IContained):
    """Level contained in an ILevelContainer."""

    containers(ILevelContainer)


class ILevelContainerContainer(IContainer):
    """Container of level containers."""

    contains(ILevelContainer)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool report pages.
"""
import cgi
import datetime
import re

try:
    import Image
except ImportError:
    from PIL import Image

from reportlab.lib import units, pagesizes

import zope.schema
from zope.component import adapts, queryMultiAdapter, getMultiAdapter
from zope.cachedescriptors.property import Lazy
from zope.interface import implements, Interface
from zope.i18n import translate
from zope.publisher.browser import BrowserView
from z3c.rml import rml2pdf

from schooltool.app import pdf
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.skin.flourish.helpers import quoteFilename
from schooltool.skin.flourish import content
from schooltool.skin.flourish import interfaces
from schooltool.skin.flourish import form
from schooltool.skin.flourish import page
from schooltool.skin.flourish import viewlet
from schooltool.skin.flourish import templates

from schooltool.common import SchoolToolMessage as _
from schooltool.common import format_message


class Box(object):

    def __init__(self, top, right=None, bottom=None, left=None):
        self.top = top
        if right is None:
            assert bottom is None, "Need to set right"
            assert left is None, "Need to set right"
            right = top
        self.right = right
        if left is None:
            self.left = right
            if bottom is None:
                self.bottom = top
            else:
                self.bottom = bottom
        else:
            assert bottom is not None, "Need to set bottom"
            self.left = left
            self.bottom = bottom


class PDFPage(page.PageBase):
    implements(interfaces.IPDFPage)

    default_content_type = 'xml'
    template = templates.XMLFile('rml/pdf.pt')
    content_template = None

    page_size = pagesizes.A4
    margin = Box(0.75*units.inch)
    rotation = 0

    title = u''
    author = _("SchoolTool")

    inline = False # Display PDF in browser? (otherwise download)

    pdf_disabled_text = _("PDF support is disabled."
                          "  It can be enabled by your administrator.")

    # TODO: Should return True when running tests
    render_invariant = False
    # TODO: Should return True when devmode enabled
    render_debug = False

    def renderPDF(self, xml):
        filename = self.filename
        stream = rml2pdf.parseString(xml, filename=filename or None)
        data = stream.getvalue()
        response = self.request.response
        response.setHeader('Content-Type', 'application/pdf')
        response.setHeader('Content-Length', len(data))
        # We don't really accept ranges, but Acrobat Reader will not show the
        # report in the browser page if this header is not provided.
        response.setHeader('Accept-Ranges', 'bytes')
        disposition = self.inline and 'inline' or 'attachment'
        quoted_filename = quoteFilename(filename)
        if quoted_filename:
            disposition += '; filename="%s"' % quoted_filename
        response.setHeader('Content-Disposition', disposition)
        return data

    @property
    def base_filename(self):
        filename = self.__name__
        if filename.strip().lower().endswith('.pdf'):
            filename = filename[:-4]
        return filename

    def makeFileName(self, basename):
        if self.render_invariant:
            return '%s.pdf' % basename
        timestamp = datetime.datetime.now().strftime('%y-%m-%d-%H-%M')
        return '%s_%s.pdf' % (basename, timestamp)

    @property
    def filename(self):
        return self.makeFileName(self.base_filename)

    def __call__(self):
        if not pdf.isEnabled():
            return translate(self.pdf_disabled_text, context=self.request)
        self.update()
        if self.request.response.getStatus() in [300, 301, 302, 303,
                                                 304, 305, 307]:
            return u''

        rml = self.render()
        data = self.renderPDF(rml)

        return data


class IPlainPDFPage(interfaces.IPDFPage):

    message_title = zope.schema.TextLine(
        title=u"Short report description", required=False)

    name = zope.schema.TextLine(
        title=u"Report name", required=False)

    scope = zope.schema.TextLine(
        title=u"Report scope (e.g. date range)", required=False)

    title = zope.schema.TextLine(
        title=u"Title", required=False)

    subtitles_left = zope.schema.Text(
        title=u"Subtitles at left", required=False)

    subtitles_right = zope.schema.TextLine(
        title=u"Subtitles at right", required=False)


class PlainPDFPage(PDFPage):
    implements(IPlainPDFPage)

    name = None
    scope = None
    title = None
    subtitles_left = None
    subtitles_right = None
    message_title = _('report')


class PDFInitSection(viewlet.ViewletManager):
    pass


class PDFPageInfoSection(viewlet.ViewletManager):

    @property
    def page_size(self):
        return self.view.page_size


class PDFStylesheetInitSection(viewlet.ViewletManager):
    pass


class PDFStylesheetSection(viewlet.ViewletManager):
    pass


class PDFTemplateSection(viewlet.ViewletManager):

    page_size = property(lambda self: self.view.page_size)
    margin = property(lambda self: self.view.margin)
    rotation = property(lambda self: self.view.rotation)
    title = property(lambda self: self.view.title)
    author = property(lambda self: self.view.author)


class PDFStory(viewlet.ViewletManager):
    pass


class PageTemplate(viewlet.Viewlet):
    implements(interfaces.IPageTemplate)

    template = None
    slots_interface = interfaces.ITemplateSlots

    @Lazy
    def slots(self):
        interface = self.slots_interface
        if interface is None:
            return None
        slots = queryMultiAdapter(
            (self.context, self.request, self.view, self), interface)
        return slots


class PageTemplateSlots(object):
    implements(interfaces.ITemplateSlots)
    adapts(Interface, interfaces.IFlourishLayer,
           interfaces.IPDFPage, interfaces.IPageTemplate)

    __name__ = 'template_slots'

    def __init__(self, context, request, view, template):
        self.__parent__ = template
        self.context = context
        self.request = request
        self.view = view
        self.template = template


class IPlainTemplateSlots(interfaces.ITemplateSlots):

    top_left = zope.schema.TextLine(title=u"Text at top bar left", required=False)
    top_center = zope.schema.TextLine(title=u"Text at top bar center", required=False)
    top_right = zope.schema.TextLine(title=u"Text at top bar right", required=False)

    title = zope.schema.TextLine(title=u"Title", required=False)
    subtitles_left = zope.schema.Tuple(
        title=u"Subtitles (left)",
        value_type=zope.schema.TextLine(title=u"subtitle", required=False),
        required=False,
        max_length=3,
        )
    subtitles_right = zope.schema.Tuple(
        title=u"Subtitles (right)",
        value_type=zope.schema.TextLine(title=u"subtitle", required=False),
        required=False,
        max_length=5,
        )


class PlainPageTemplate(PageTemplate):

    slots_interface = IPlainTemplateSlots
    header_padding_top = 4
    header_padding_bottom = 8
    min_header_lines = 5

    def lines(self, attr, top, left):
        content = attr['content']
        if not isinstance(content, list):
            content = [content]
        content = filter(None, content)
        margin = attr['margin']
        result = []
        line_height = attr['fontSize'] + margin.top + margin.bottom
        for i, line in enumerate(content):
            result.append({
                    'content': line,
                    'x': left,
                    'y': top - (line_height * (i+1)) + margin.bottom,
                    })
        return line_height * len(result), result

    @Lazy
    def top_line(self):
        width = self.frame['width']
        x1 = self.frame['x']
        x2 = x1 + width
        y = self.header['y'] - (self.frame['margin'].top / 2.0)
        return {
            'coords': '%d %d %d %d' % (x1, y, x2, y),
            }

    @Lazy
    def bottom_line(self):
        width = self.frame['width']
        x1 = self.frame['x']
        x2 = x1 + width
        y = self.frame['y'] - (self.frame['margin'].top / 2.0)
        return {
            'coords': '%d %d %d %d' % (x1, y, x2, y),
            }

    @Lazy
    def header_title(self):
        title = {
            'fontSize': 20,
            'margin': Box(2, 0, 6, 0),
            'content': self.slots.title,
            }
        top = self.top_bar['y'] - self.header_padding_top
        left = self.manager.margin.left
        title['height'], title['lines'] = self.lines(title, top, left)
        return title

    @Lazy
    def header_subtitles(self):
        doc_w, doc_h = self.manager.page_size
        subtitles = {
            'fontSize': 12,
            'margin': Box(1, 0, 1, 0),
            'content': self.slots.subtitles_left,
            }

        top = self.top_bar['y'] - self.header_padding_top - self.header_title['height']
        left = self.manager.margin.left
        subtitles['height'], subtitles['lines'] = self.lines(
            subtitles, top, left)
        return subtitles

    @Lazy
    def header_extra_subtitles(self):
        extra_subtitles = {
            'fontSize': 12,
            'margin': Box(1, 0, 1, 0),
            'content': self.slots.subtitles_right,
            }

        if not self.slots.subtitles_right:
            prefs = IApplicationPreferences(ISchoolToolApplication(None))
            extra_subtitles['content'] = prefs.title

        doc_w, doc_h = self.manager.page_size
        top = self.top_bar['y'] - self.header_padding_top
        right = doc_w - self.manager.margin.right
        extra_subtitles['height'], extra_subtitles['lines'] = self.lines(
            extra_subtitles, top, right)
        return extra_subtitles

    @Lazy
    def header_school_logo(self):
        if self.slots.subtitles_right:
            return None

        prefs = IApplicationPreferences(ISchoolToolApplication(None))
        if prefs.logo is None:
            return None

        padding = Box(4, 4, 0, 4)

        # XXX: need a better way to get image size
        logo_file = prefs.logo.open()
        image = Image.open(logo_file)
        logo_file.close()

        doc_w, doc_h = self.manager.page_size
        top = self.top_bar['y'] - self.header_padding_top
        right = doc_w - self.manager.margin.right

        extra_subtitles = self.header_extra_subtitles
        height = max(
            self.header_title['height'] + self.header_subtitles['height'],
            (extra_subtitles['fontSize'] +
             extra_subtitles['margin'].top +
             extra_subtitles['margin'].bottom) * self.min_header_lines)
        height += self.header_padding_top + self.header_padding_bottom
        height -= extra_subtitles['height'] + padding.top + padding.bottom
        ratio = image.size and float(image.size[0])/image.size[1] or 1
        width = height*ratio

        logo_data = getMultiAdapter((prefs.logo, self.request),
                                    name='data_uri')

        logo = {
            'x': right - width - padding.right,
            'y': top - extra_subtitles['height'] - height - padding.top,
            'width': width,
            'height': height,
            'logo_data': logo_data,
            }
        return logo

    @Lazy
    def header(self):
        doc_w, doc_h = self.manager.page_size
        title = self.header_title
        subtitles = self.header_subtitles
        extra_subtitles = self.header_extra_subtitles
        height = max(
            title['height'] + subtitles['height'],
            extra_subtitles['height'],
            (subtitles['fontSize'] +
             subtitles['margin'].top +
             subtitles['margin'].bottom) * self.min_header_lines,
            (extra_subtitles['fontSize'] +
             extra_subtitles['margin'].top +
             extra_subtitles['margin'].bottom) * self.min_header_lines)

        width = doc_w - self.manager.margin.left - self.manager.margin.right
        x = self.manager.margin.left
        height += self.header_padding_top + self.header_padding_bottom
        y = self.top_bar['y'] - height -1
        logo = self.header_school_logo
        header = {
            'title': title,
            'subtitle': subtitles,
            'extra_subtitle': extra_subtitles,
            'logo': logo,
            'height': height,
            'width': width,
            'x': x,
            'y': y,
            }
        return header

    @Lazy
    def top_bar(self):
        fontSize = 12
        padding = Box(5-fontSize/6, 10.5, 5+fontSize/6, 10.5)
        height = fontSize + padding.top + padding.bottom
        doc_w, doc_h = self.manager.page_size
        width = doc_w - self.manager.margin.left - self.manager.margin.right
        x = self.manager.margin.left
        y = doc_h - self.manager.margin.top - height
        slot_y = doc_h - self.manager.margin.top - padding.top - fontSize
        bar = {
            'height': height,
            'width': width,
            'x': x,
            'y': y,
            'fontSize': fontSize,
            'slots': {
                'left': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'center': {
                    'x': x + width/2,
                    'y': slot_y,
                    },
                'right': {
                    'x': x + width - padding.right,
                    'y': slot_y,
                    },
                }
            }
        return bar

    @Lazy
    def bottom_bar(self):
        fontSize = 8.5
        padding = Box(1.5, 10.5)
        height = fontSize + padding.top + padding.bottom
        doc_w, doc_h = self.manager.page_size
        width = doc_w - self.manager.margin.left - self.manager.margin.right
        x = self.manager.margin.left
        y = self.manager.margin.bottom
        slot_y = y + padding.bottom
        return {
            'fontSize': fontSize,
            'height': height,
            'width': width,
            'x': x,
            'y': y,
            'slots': {
                'left': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'center': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'right': {
                    'x': x + width - padding.right,
                    'y': slot_y,
                    },
                }
            }

    @Lazy
    def frame(self):
        doc_w, doc_h = self.manager.page_size
        margin = Box(8, 0)
        width = (doc_w - self.manager.margin.left - self.manager.margin.right
                 - margin.left - margin.right)
        height = (doc_h - self.manager.margin.top - self.manager.margin.bottom
                  - self.top_bar['height'] - self.bottom_bar['height']
                  - self.header['height']
                  - margin.top - margin.bottom)
        x = self.manager.margin.left + margin.left
        y = (self.manager.margin.bottom + self.bottom_bar['height'] +
             margin.bottom)
        return {
            'height': height,
            'margin': margin,
            'width': width,
            'x': x,
            'y': y,
            }


class PlainPageTemplateSlots(PageTemplateSlots):
    adapts(Interface, interfaces.IFlourishLayer,
           interfaces.IPDFPage, PlainPageTemplate)
    implements(IPlainTemplateSlots)

    @property
    def top_left(self):
        return self.view.name.upper()

    top_center = None

    @property
    def top_right(self):
        return self.view.scope

    @property
    def title(self):
        return self.view.title

    @property
    def subtitles_left(self):
        subtitles = self.view.subtitles_left or []
        if (subtitles and
            not isinstance(subtitles, (list, tuple))):
            subtitles = translate(subtitles, context=self.request)
            subtitles = [s.strip() for s in subtitles.strip().splitlines()]
        subtitles = subtitles[:IPlainTemplateSlots['subtitles_left'].max_length]
        return subtitles

    @property
    def subtitles_right(self):
        subtitles = self.view.subtitles_right or []
        if (subtitles and
            not isinstance(subtitles, (list, tuple))):
            subtitles = translate(subtitles, context=self.request)
            subtitles = [s.strip() for s in subtitles.strip().splitlines()]
        subtitles = subtitles[:IPlainTemplateSlots['subtitles_right'].max_length]
        return subtitles


class PDFContentBlock(content.ContentProvider):

    title = None

    @property
    def title_continued(self):
        if not self.title:
            return None
        text = translate(
            format_message(
                _("${title} (continued)"),
                mapping={'title': self.title}),
            context=self.request)
        return text

    @property
    def templates(self):
        return self.view.providers.get('template')


class PDFPart(viewlet.Viewlet, PDFContentBlock):
    implements(interfaces.IPDFPart)


class RMLWidgetRows(content.ContentProvider):
    implements(interfaces.IRMLTemplated)

    @property
    def hidden(self):
        return self.context.mode == 'hidden' or not self.context.value

    def rows(self):
        return [self.context]


class RMLMultiWidgetRows(RMLWidgetRows):

    def rows(self):
        result = []
        for widget in self.context.widgets:
            row = content.queryContentProvider(
                widget, self.request, self.view, self.__name__)
            if row is not None:
                result.extend(row.rows())
        return result


class PDFForm(PDFPart, form.DisplayForm):

    template = templates.XMLFile('rml/pdf_form.pt')

    def update(self):
        form.DisplayForm.update(self)
        PDFPart.update(self)

    def render(self, *args, **kw):
        return form.DisplayForm.render(self, *args, **kw)


def text2rml(snippet, para_class=None):
    if not snippet:
        return snippet
    snippet = unicode(snippet)
    paragraphs = filter(None, [p.strip() for p in snippet.splitlines()])
    if not paragraphs:
        return ''
    start_tag = '<para>'
    if para_class is not None:
        start_tag = '<para style="%s">' % para_class
    return start_tag+('</para>\n'+start_tag).join(paragraphs)+'</para>'


def unescapeHTMLEntities(text):
    text = text.replace(u'&amp;', u'&')
    text = text.replace(u'&lt;', u'<')
    text = text.replace(u'&gt;', u'>')
    text = text.replace(u'&quot;', u'"')
    text = text.replace(u'&#39;', u"'")
    text = text.replace(u'&rsquo;', u"'")
    text = text.replace(u'&nbsp;', u' ')
    return text


html_tags_valid_in_rml_re = re.compile(
    r'&lt;(/?((strong)|(b)|(em)|(i)))&gt;')

html_p_tag_re = re.compile(r'</?p[^>]*>')
html_br_tag_re = re.compile(r'</?br[^>]*>')


def buildHTMLParagraphs(snippet):
    """Build a list of paragraphs from an HTML snippet."""
    if not snippet:
        return []
    paragraphs = []
    tokens = []
    for token in html_p_tag_re.split(snippet):
        if not token or token.isspace():
            continue
        tokens.extend(html_br_tag_re.split(token))
    for token in tokens:
        if not token or token.isspace():
            continue
        # Reportlab is very sensitive to unknown tags and escaped symbols.
        # In case of invalid HTML, try to provide correct escaping.
        fixed_escaping = cgi.escape(unescapeHTMLEntities(unicode(token)))
        # Unescape some of the tags which are also valid in Reportlab
        valid_text = html_tags_valid_in_rml_re.sub(u'<\g<1>>', fixed_escaping)
        paragraphs.append(valid_text)
    return paragraphs


def html2rml(snippet, para_class=None):
    paragraphs = buildHTMLParagraphs(snippet)
    if not paragraphs:
        return ''
    start_tag = '<para>'
    if para_class is not None:
        start_tag = '<para style="%s">' % para_class
    return start_tag+('</para>\n'+start_tag).join(paragraphs)+'</para>'


class Text2RML(BrowserView):

    para_class = None

    def __call__(self):
        text = self.context
        if not text:
            return text
        snippet = translate(text, context=self.request)
        rml = text2rml(snippet, self.para_class)
        return rml


class HTML2RML(BrowserView):

    para_class = None

    def __call__(self):
        text = self.context
        if not text:
            return text
        snippet = translate(text, context=self.request)
        rml = html2rml(snippet, self.para_class)
        return rml


class PDFPage2RML(BrowserView):

    def __call__(self):
        self.context.update()
        rml = self.context.render()
        return rml


class Image2RML(BrowserView):

    template = templates.Inline('''
      <tal:block xmlns:tal="http://xml.zope.org/namespaces/tal"
        ><imageAndFlowables tal:attributes="imageName context/@@data_uri"
        ></imageAndFlowables
      ></tal:block>
    ''', content_type="text/xml")

    def __call__(self, *args, **kw):
        return self.template()
